// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import net.orfjackal.util.*;
import org.jetbrains.annotations.*;

import java.io.*;
import java.net.URL;
import java.util.*;

/**
 * Creates reports of the provided source code files.
 *
 * @author Esko Luontola
 */
public class AnalyzerBatch {

    private static final String[] PROFILER_CLASSES = new String[]{
            "net/orfjackal/sorsastus/ProfilerLogger$1.class",
            "net/orfjackal/sorsastus/ProfilerLogger$ThreadState.class",
            "net/orfjackal/sorsastus/ProfilerLogger.class"
    };
    private static final String PROFILER_DIRECTORY = "net/orfjackal/sorsastus";
    private static final String PROFILER_JAVA_FILE_REGEX = ".*?net.orfjackal.sorsastus.ProfilerLogger\\.java\\z";

    @NotNull
    private List<FileAnalyzer> files;

    @NotNull
    private Set<File> classpaths;

    @Nullable
    private List<FileAnalyzer> classpathFileCache;

    private boolean countCodeLines;

    private boolean countCommentLines;

    private boolean countEmptyLines;

    private boolean includeOOMetrics;

    @Nullable
    private File profilerJDKPath;

    @Nullable
    private String profilerMainClass;

    @NotNull
    private List<String> profilerParameters;

    private int profilerRuns;

    @NotNull
    private List<ProfilerLogParser> profilers;

    public AnalyzerBatch() {
        files = new ArrayList<FileAnalyzer>();
        classpaths = new HashSet<File>();
        classpathFileCache = null;
        countCodeLines = true;
        countCommentLines = false;
        countEmptyLines = false;
        includeOOMetrics = false;
        profilerJDKPath = null;
        profilerMainClass = null;
        profilerParameters = new ArrayList<String>();
        profilerRuns = 1;
        profilers = new ArrayList<ProfilerLogParser>();
    }

    /**
     * Adds a file to the list files being analyzed. Will read contents of the file to memory.
     *
     * @param file a source code file.
     * @throws NullPointerException if any parameter is null.
     * @throws IOException          if the parameter does not exist, is not a file or can not be read. The message of
     *                              the exception will contain the explanation.
     */
    public void addFile(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException();
        } else if (!file.exists()) {
            throw new IOException("No such file: " + file);
        } else if (!file.isFile()) {
            throw new IOException("Not a file: " + file);
        } else if (!file.canRead()) {
            throw new IOException("File can not be read: " + file);
        }
        FileAnalyzer fa = new FileAnalyzer(this, file);
        files.add(fa);
        classpathFileCache = null;
    }

    /**
     * Updates the classpath list by extracting the root folder from the files being analyzed.
     */
    private void updateClasspaths() {

        // extract the classpaths from the files
        for (FileAnalyzer fa : files) {
            String[] dirs = fa.getPackage().split("\\.");
            if (fa.getPackage().length() == 0) {
                // otherwise classes in default package will be placed in a subdirectory
                dirs = new String[0];
            }
            try {
                File f = fa.getFile().getCanonicalFile();
                for (int i = 0; i < dirs.length + 1 && f != null; i++) {
                    f = f.getParentFile();
                }
                if (f != null) {
                    classpaths.add(f);      // the Set will remove all duplicates
                }
            } catch (IOException e) {
                if (Tools.isDebug()) e.printStackTrace();
            }
        }
    }

    /**
     * Returns an array of the directories in the classpath. Updates the classpath list from the files that are being
     * analyzed.
     */
    public File[] getClasspaths() {
        updateClasspaths();
        return classpaths.toArray(new File[classpaths.size()]);
    }

    public boolean isCountCodeLines() {
        return countCodeLines;
    }

    public void setCountCodeLines(boolean countCodeLines) {
        this.countCodeLines = countCodeLines;
    }

    public boolean isCountCommentLines() {
        return countCommentLines;
    }

    public void setCountCommentLines(boolean countCommentLines) {
        this.countCommentLines = countCommentLines;
    }

    public boolean isCountEmptyLines() {
        return countEmptyLines;
    }

    public void setCountEmptyLines(boolean countEmptyLines) {
        this.countEmptyLines = countEmptyLines;
    }

    public boolean isIncludeOOMetrics() {
        return includeOOMetrics;
    }

    public void setIncludeOOMetrics(boolean includeOOMetrics) {
        this.includeOOMetrics = includeOOMetrics;
    }

    @Nullable
    public File getProfilerJDKPath() {
        return profilerJDKPath;
    }

    public void setProfilerJDKPath(File profilerJDKPath) {
        this.profilerJDKPath = profilerJDKPath;
    }

    @Nullable
    public String getProfilerMainClass() {
        return profilerMainClass;
    }

    public void setProfilerMainClass(String profilerMainClass) {
        this.profilerMainClass = profilerMainClass;
    }

    public void addProfilerParameter(String param) {
        if (param == null) {
            throw new NullPointerException("param is null");
        }
        this.profilerParameters.add(param);
    }

    public boolean isProfilerEnabled() {
        return profilerMainClass != null;
    }

    public int getProfilerRuns() {
        return profilerRuns;
    }

    public void setProfilerRuns(int profilerRuns) {
        if (profilerRuns >= 1) {
            this.profilerRuns = profilerRuns;
        }
    }

    @NotNull
    public List<ProfilerLogParser> getProfilers() {
        return profilers;
    }

    /**
     * Returns a report which contains the line counts of each of the input files. For each file is reported how many
     * lines of code it includes (according to the set LOC definition) and the name of the file. The total sum of the
     * line counts is told on the last line.
     * <p/>
     * The report will end with the line freed character '\n'.
     */
    @NotNull
    public String getLineCountReport() {
        StringBuilder sb = new StringBuilder();
        int totalLines = 0;
        for (FileAnalyzer file : files) {
            int lines = getLineCount(file);
            totalLines += lines;
            sb.append(lines);
            sb.append('\t');
            sb.append(file.getFile().getPath());
            sb.append('\n');
        }
        sb.append(totalLines);
        sb.append("\ttotal\n");
        return sb.toString();
    }

    private int getLineCount(FileAnalyzer file) {
        int count = 0;
        if (isCountCodeLines()) {
            count += file.getCodeLines();
        }
        if (isCountCommentLines()) {
            count += file.getCommentLines();
        }
        if (isCountEmptyLines()) {
            count += file.getEmptyLines();
        }
        return count;
    }

    /**
     * Runs the profiler according to the settings that have been set in this class. After this is complete, the
     * structure report will include the runtime metrics of the profiled program. If there is an error parsing the
     * profiler code, this program will exit with System.exit(1).
     * <p/>
     * The program will be executed in the current working directory. The temporary source code and class files will be
     * automatically deleted if the program was profiled successfully.
     */
    public void doProfiler() {
        if (!isProfilerEnabled()) {
            throw new IllegalStateException("Profiler is not enabled");
        }

        // make a new directory for all temporary files
        System.out.println("-- Preparing files --");
        File tempDir = new File(".profilerTemp");
        if (tempDir.exists()) {
            int i = 0;
            do {
                i++;
                tempDir = new File(".profilerTemp-" + i);
            } while (tempDir.exists());
        }
        if (!tempDir.mkdir()) {
            System.err.println("Unable to create temporary directory: " + tempDir);
            System.exit(1);
        }

        // copy profiled source code to the temporary directory, make a clean copy (remove all .class files)
        insertProfilerCode(tempDir);
        List<File> sourceCodeFiles = new ArrayList<File>();

        Stack<File> tempDirFiles = new Stack<File>();
        tempDirFiles.push(tempDir);
        while (tempDirFiles.size() > 0) {
            File file = tempDirFiles.pop();
            if (file.isDirectory()) {
                for (File f : file.listFiles()) {
                    tempDirFiles.push(f);
                }
            } else if (file.getName().matches(".*?\\.class\\z")) {
                file.delete();
            } else if (file.getName().matches(".*?\\.java\\z")) {
                sourceCodeFiles.add(file);
            }
        }

        // compile the source code
        System.out.println("-- Compiling source code --");
        List<String> compileCommand = new ArrayList<String>();
        if (profilerJDKPath == null) {
            compileCommand.add("javac");
        } else {
            compileCommand.add(profilerJDKPath + System.getProperty("file.separator")
                    + "bin" + System.getProperty("file.separator") + "javac");
        }
        compileCommand.add("-sourcepath");
        compileCommand.add(tempDir.getPath());
        compileCommand.add("-classpath");
        compileCommand.add(System.getProperty("java.class.path")); // use the local JVM's classpath
        for (File file : sourceCodeFiles) {
            compileCommand.add(file.getPath());
        }
        try {
            Process p = Runtime.getRuntime().exec(compileCommand.toArray(new String[0]));

            StreamGobbler in = new StreamGobbler(
                    new BufferedInputStream(p.getInputStream()),
                    System.out
            );
            StreamGobbler err = new StreamGobbler(
                    new BufferedInputStream(p.getErrorStream()),
                    System.err
            );
            in.start();
            err.start();
            int returnval = p.waitFor();
            in.close();
            err.close();

            if (returnval != 0) {
                System.err.println("Compiler finished with exit code " + returnval);
                System.exit(1);
            }
        } catch (Exception e) {
            if (Tools.isDebug()) e.printStackTrace();
            System.err.println("Error in compiling the source code: " + e.getMessage());
            System.exit(1);
        }

        // copy the class files needed by the profiler to the temporary directory
        new File(tempDir, PROFILER_DIRECTORY).mkdirs();
        for (String classFile : PROFILER_CLASSES) {
            URL url = ClassLoader.getSystemResource(classFile);
            try {
                InputStream in = (InputStream) url.getContent(new Class[]{InputStream.class});
                FileOutputStream out = new FileOutputStream(new File(tempDir, classFile));


                int c;
                while ((c = in.read()) != -1) {
                    out.write(c);
                }
                out.close();
            } catch (IOException e) {
                if (Tools.isDebug()) e.printStackTrace();
                System.err.println("Error copying " + classFile + " to temporary directory: " + e.getMessage());
                System.exit(1);
            }
        }

        // executing the profiler and analyzing the logs
        List<String> runCommand = new ArrayList<String>();
        if (profilerJDKPath == null) {
            runCommand.add("java");
        } else {
            runCommand.add(profilerJDKPath + System.getProperty("file.separator")
                    + "bin" + System.getProperty("file.separator") + "java");
        }
        runCommand.add("-classpath");
        runCommand.add(tempDir.getPath() + ";" + System.getProperty("java.class.path"));
//        runCommand.add(tempDir.getPath());
        runCommand.add(profilerMainClass);
        for (String param : profilerParameters) {
            runCommand.add(param);
        }
        for (int i = 0; i < profilerRuns; i++) {

            // find out which log files are from previous runs, so that we can ignore them in the analyzing part
            FilenameFilter logFileFilter = new LogFilenameFilter();
            final File[] oldLogFiles = ProfilerLogger.LOG_DIRECTORY.listFiles(logFileFilter);

            // run the profiler
            try {
                System.out.println("-- Profiler run " + (i + 1) + " of " + profilerRuns + " --");
                Process p = Runtime.getRuntime().exec(runCommand.toArray(new String[0]));
                StreamGobbler in = new StreamGobbler(
                        new BufferedInputStream(p.getInputStream()),
                        new PrintStream(new FileOutputStream(new File(tempDir, "stdout-" + i + ".log")))
                );
                StreamGobbler err = new StreamGobbler(
                        new BufferedInputStream(p.getErrorStream()),
                        new PrintStream(new FileOutputStream(new File(tempDir, "stderr-" + i + ".log")))
                );
                in.start();
                err.start();
                p.waitFor();
                in.close();
                err.close();
            } catch (Exception e) {
                if (Tools.isDebug()) e.printStackTrace();
            }

            // analyze the profiler logs, ignore logs from previous runs
            System.out.println("-- Analyzing logs --");
            List<File> logFiles = new ArrayList<File>();
            logLoop:
            for (File file : ProfilerLogger.LOG_DIRECTORY.listFiles(logFileFilter)) {
                for (File oldFile : oldLogFiles) {
                    if (file.equals(oldFile)) {
                        continue logLoop;
                    }
                }
                logFiles.add(file);
            }
            try {
                profilers.add(new ProfilerLogParser(logFiles.toArray(new File[0])));
            } catch (IOException e) {
                if (Tools.isDebug()) e.printStackTrace();
                System.err.println("Unable to read profiler logs: " + e.getMessage());
                System.exit(1);
            }
        }

        // remove temporary files
        System.out.println("-- Cleanup --");
        FileTools.recursiveDelete(tempDir);
    }

    // HACK: needed because of the profiler, which will generate bad code for anonymous inner classes containing return/catch statements
    private class LogFilenameFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            return name.matches(".*?\\." + ProfilerLogParser.LOG_FILE_TYPE + "\\z");
        }
    }

    private class StreamGobbler extends Thread {

        private InputStream in;
        private PrintStream out;

        public StreamGobbler(InputStream in, PrintStream out) {
            this.in = in;
            this.out = out;
        }

        @Override
        public void run() {
            try {
                int c;
                while ((c = in.read()) != -1) {
                    if (out != null) {
                        out.print((char) c);
                    }
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        public void close() {
            this.interrupt();
            if (out == System.out || out == System.err) {
                out.flush();
            } else {
                out.close();
            }
        }
    }

    /**
     * Returns a report which contains a structural analysis of each of the input files. The report will contain
     * <ul><li>the name of the source code file(s) processed</li> <li>for each source file:</li> <ul><li>the name of the
     * package the file belongs to,</li><li>the number and names of packages imported,</li><li>the number and names of
     * classes in the file,</li></ul> <li>for each class:</li> <ul><li>the number and names of the methods in the
     * class,</li></ul> <li>for each method:</li> <ul><li>the number of if and switch statements,</li><li>the number of
     * for, while and do-while statements.</li></ul></ul> The report will end with the line freed character '\n'.
     *
     * @throws ANTLRException if there was a parse exception in processing one of the files.
     */
    @NotNull
    public String getStructureReport() throws ANTLRException {
        StringBuilder sb = new StringBuilder();

        if (getProfilers().size() > 0) {
            sb.append("Number of Profiler Executions: ").append(getProfilers().size()).append("\n");
        }

        for (FileAnalyzer file : files) {
            sb.append(file.getFile().getPath());
            sb.append('\n');
            try {
                sb.append(file.getStructureReport());
                sb.append('\n');
            } catch (ANTLRException e) {
                if (Tools.isDebug()) e.printStackTrace();
                throw new ANTLRException("Parse exception in file '" + file.getFile() + "': " + e.getMessage());
            }
        }
        return sb.toString();
    }

    /**
     * Tries to find a the specified class from the classpath.
     *
     * @param classPackage the name of the package of the class.
     * @param className    the name of the class.
     * @return the associated ClassAnalyzer, or null if it was not found.
     * @throws NullPointerException if any argument is null.
     */
    @Nullable
    public ClassAnalyzer findClass(String classPackage, String className) {
//        System.out.println("findClass: " + classPackage + "." + className); // DEBUG
        if (classPackage == null) {
            throw new NullPointerException();
        } else if (className == null) {
            throw new NullPointerException();
        }

        // look for the class in the files being analyzed
        for (FileAnalyzer fa : files) {
            ClassAnalyzer ca = fa.getClass(classPackage, className);
            if (ca != null) {
                return ca;
            }
        }

        // look for the class in other files in the classpath
        String fileSeparator = System.getProperty("file.separator");
        String classFile = classPackage.replace(".", fileSeparator) + fileSeparator + className + ".java";
        for (File classpath : getClasspaths()) {
            File f = new File(classpath, classFile);
            if (f.exists() && f.isFile() && f.canRead()) {
                try {
                    FileAnalyzer fa = new FileAnalyzer(this, f);
                    ClassAnalyzer ca = fa.getClass(classPackage, className);
                    if (ca != null) {
                        return ca;
                    }
                } catch (IOException e) {
                    if (Tools.isDebug()) e.printStackTrace();
                }
            }
        }
        return null;
    }

    @NotNull
    public List<FileAnalyzer> getClasspathFileCache() {
        if (classpathFileCache == null) {
            ArrayList<FileAnalyzer> result = new ArrayList<FileAnalyzer>();

            // go through all files in the classpaths
            List<File> files = new ArrayList<File>();
            for (File classpath : getClasspaths()) {
                FileTools.recursiveAdd(files, classpath);
            }
            for (File file : files) {

                // ignore non-Java files
                if (!file.getName().matches(".*?\\.java\\z")) {
                    continue;
                }

                // analyze all Java files and add them to the cache
                try {
                    result.add(new FileAnalyzer(this, file));
                } catch (IOException e) {
                    if (Tools.isDebug()) e.printStackTrace();
                }
            }
            classpathFileCache = result;
        }
        return classpathFileCache;
    }

    /**
     * Tries to find from the classpath all classes that extend or implement the specified class. Only classes that
     * inherit it directly are returned.
     *
     * @param parentClass the parent class, whose direct children we are looking for.
     * @return an array of the classes that were found. If none were found, the size of the array is zero.
     * @throws NullPointerException if any argument is null.
     */
    @NotNull
    public ClassAnalyzer[] findChildClasses(ClassAnalyzer parentClass) {
        if (parentClass == null) {
            throw new NullPointerException();
        }
        List<ClassAnalyzer> result = new ArrayList<ClassAnalyzer>();
        List<FileAnalyzer> files = getClasspathFileCache();

        // find the classes that extend or implement parentClass
        for (FileAnalyzer fa : files) {
            for (ClassAnalyzer ca : fa.getClasses()) {
                ClassAnalyzer caExtends = ca.getExtends();
                ClassAnalyzer[] caImplements = ca.getImplements();
                boolean isChild = false;
                if (caExtends != null && caExtends.equals(parentClass)) {
                    isChild = true;
                }
                for (ClassAnalyzer impl : caImplements) {
                    if (impl.equals(parentClass)) {
                        isChild = true;
                    }
                }
                if (isChild) {
                    result.add(ca);         // is a child of parentClass, add to results
                }
            }
        }
        return result.toArray(new ClassAnalyzer[result.size()]);
    }

    /**
     * Inserts the profiler code to the source code files. Copies the modified files and all other files in classpath to
     * outputDir. The contents of the outputDir will be overwritten. Will not modify the original files.
     */
    public void insertProfilerCode(File outputDir) {

        // copy all files in classpath to the output dir
        File[] classPaths = getClasspaths();
        for (File classPath : classPaths) {
            try {
                File[] files = classPath.listFiles();
                for (File f : files) {
                    FileTools.recursiveCopy(f, outputDir, true);
                }
            } catch (IOException e) {
                if (Tools.isDebug()) e.printStackTrace();
                System.err.println("Unable to copy files: " + e.getMessage());
                System.exit(1);
            }
        }

        for (FileAnalyzer fa : files) {

            // do not profile the profiler itself
            if (fa.getFile().getAbsolutePath().matches(PROFILER_JAVA_FILE_REGEX)) {
                continue;
            }

            String code = "";
            try {
                code = fa.getProfilerCode();

                // extract the file path relative to classpath, and write profiled code to output dir
                for (File classPath : classPaths) {
                    String srcPath = fa.getFile().getCanonicalPath();
                    String clsPath = classPath.getCanonicalPath();
                    if (!srcPath.startsWith(clsPath)) {
                        continue;
                    }
                    String relPath = srcPath.substring(clsPath.length() + 1);
                    File outFile = new File(outputDir, relPath);

                    FileWriter writer = new FileWriter(outFile);
                    writer.append(code);
                    writer.close();
                }

            } catch (ANTLRException e) {
                if (Tools.isDebug()) e.printStackTrace();
                System.err.println("Unable parse file: " + fa.getFile());
                System.exit(1);

            } catch (IOException e) {
                if (Tools.isDebug()) e.printStackTrace();
                System.err.println("IO Error: " + e.getMessage());
                System.exit(1);
            }
        }
    }
}
