// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import antlr.collections.AST;
import net.orfjackal.util.*;
import org.jetbrains.annotations.*;
import util.java15grammar.JavaTreeParserTokenTypes;

import java.math.BigInteger;
import java.util.*;

/**
 * Collects metrics from the structure of a class.
 *
 * @author Esko Luontola
 */
public class ClassAnalyzer {

    @NotNull
    private FileAnalyzer fileAnalyzer;

    @NotNull
    private AST classNode;

    /* Analyzed data from this class. Initialized by analyzeStructure() */
    @NotNull
    private String className;
    @Nullable
    private String classExtends;
    @NotNull
    private String[] classImplements;
    @NotNull
    private MethodAnalyzer[] classMethods;

    /**
     * Creates a new ClassAnalyzer and analyzes the structure of the class. Is not (yet) able to analyze anonymous
     * classes.
     *
     * @param fileAnalyzer the file that this class belongs to.
     * @param classNode    structure of the class. The type of the node must be CLASS_DEF or INTERFACE_DEF.
     * @throws NullPointerException     if any parameter is null.
     * @throws IllegalArgumentException if classNode is of wrong type.
     */
    public ClassAnalyzer(FileAnalyzer fileAnalyzer, AST classNode) {
        if (fileAnalyzer == null) {
            throw new NullPointerException();
        } else if (classNode == null) {
            throw new NullPointerException();
        } else if (classNode.getType() != JavaTreeParserTokenTypes.CLASS_DEF
                && classNode.getType() != JavaTreeParserTokenTypes.INTERFACE_DEF) {
            throw new IllegalArgumentException("classNode.getType() = " + classNode.getType());
        }
        this.fileAnalyzer = fileAnalyzer;
        this.classNode = classNode;
        analyzeStructure();
    }

    /**
     * Analyzes the AST tree of this class and saves the results for later use.
     */
    private void analyzeStructure() {

        // components of the class definition
        @NotNull AST cModifiers = null;
        @NotNull AST cIdent = null;
        @NotNull AST cExtendsClause = null;
        @Nullable AST cImplementsClause = null;     // null if interface class
        @NotNull AST cObjBlock = null;

        // find the components of the class definition
        AST n = classNode.getFirstChild();
        while (n != null) {
            switch (n.getType()) {
            case JavaTreeParserTokenTypes.MODIFIERS:
                cModifiers = n;
                break;
            case JavaTreeParserTokenTypes.IDENT:
                cIdent = n;
                break;
            case JavaTreeParserTokenTypes.EXTENDS_CLAUSE:
                cExtendsClause = n;
                break;
            case JavaTreeParserTokenTypes.IMPLEMENTS_CLAUSE:
                cImplementsClause = n;
                break;
            case JavaTreeParserTokenTypes.OBJBLOCK:
                cObjBlock = n;
                break;
            default:
                assert false : "n.getType() = " + n.getType();
                break;
            }
            n = n.getNextSibling();
        }
        assert cModifiers != null;
        assert cIdent != null;
        assert cExtendsClause != null;
        assert cObjBlock != null;

        // name of the class
        this.className = cIdent.getText();

        // classes extended by the class
        if (cExtendsClause.getFirstChild() != null) {
            this.classExtends = ASTTools.nodeToString(cExtendsClause.getFirstChild());
        } else {
            this.classExtends = null;
        }

        // interfaces implemented by the class
        List<String> interfaces = new ArrayList<String>();
        if (cImplementsClause != null) {
            n = cImplementsClause.getFirstChild();
            while (n != null) {
                interfaces.add(ASTTools.nodeToString(n));
                n = n.getNextSibling();
            }
        }
        this.classImplements = interfaces.toArray(new String[interfaces.size()]);

        // methods of the class; avoid the counting of methods of anonymous and private classes
        List<AST> methodNodes = new ArrayList<AST>();
        ASTTools.findNodes(
                cObjBlock,
                JavaTreeParserTokenTypes.CTOR_DEF,
                methodNodes,
                JavaTreeParserTokenTypes.CLASS_DEF,
                JavaTreeParserTokenTypes.ENUM_DEF,
                JavaTreeParserTokenTypes.CTOR_DEF,
                JavaTreeParserTokenTypes.METHOD_DEF
        );
        ASTTools.findNodes(
                cObjBlock,
                JavaTreeParserTokenTypes.METHOD_DEF,
                methodNodes,
                JavaTreeParserTokenTypes.CLASS_DEF,
                JavaTreeParserTokenTypes.ENUM_DEF,
                JavaTreeParserTokenTypes.CTOR_DEF,
                JavaTreeParserTokenTypes.METHOD_DEF
        );

        List<MethodAnalyzer> list = new ArrayList<MethodAnalyzer>(methodNodes.size());
        for (AST node : methodNodes) {
            list.add(new MethodAnalyzer(this, node));
        }
        this.classMethods = list.toArray(new MethodAnalyzer[list.size()]);
    }

    /**
     * Returns a detailed report of the structure of this class. The returned string will end with the line feed
     * character '\n'.
     */
    @NotNull
    public String getStructureReport() {
        StringBuilder sb = new StringBuilder();

        // name of the class
        sb.append('\t').append(className).append('\n');

        // object-oriented metrics of the class
        if (fileAnalyzer.getAnalyzerBatch().isIncludeOOMetrics()) {
            sb.append("\t    WMC: ").append(getWMC()).append("\t(weighted methods per class)\n");
            sb.append("\t    DIT: ").append(getDIT()).append("\t(depth of inheritance tree)\n");
            sb.append("\t    NOC: ").append(getNOC()).append("\t(number of children)\n");
        }

        // runtime report
        List<ProfilerLogParser> profilers = fileAnalyzer.getAnalyzerBatch().getProfilers();
        String classID = getPackage() + "." + getName();
        if (profilers.size() > 0) {
            BigInteger min = profilers.get(0).getClassStats(classID).getTotalExecutionTime();
            BigInteger max = min;
            for (ProfilerLogParser profiler : profilers) {
                BigInteger execTime = profiler.getClassStats(classID).getTotalExecutionTime();
                if (execTime.compareTo(min) < 0) {
                    min = execTime;
                }
                if (execTime.compareTo(max) > 0) {
                    max = execTime;
                }
            }
            sb.append("\t    Min Execution Time: ").append(ProfilerLogParser.nanoToMilli(min)).append(" ms\n");
            sb.append("\t    Max Execution Time: ").append(ProfilerLogParser.nanoToMilli(max)).append(" ms\n");
        }

        // methods of the class
        sb.append("\t    Methods (total of ").append(classMethods.length).append("):\n");
        for (MethodAnalyzer method : classMethods) {
            sb.append("\t\t").append(method.getDefinition());
            sb.append('\n');
            sb.append("\t\t    Complexity: ").append(method.getComplexity());
            sb.append("\t(Conditionals: ").append(method.getConditionals());
            sb.append(", Loops: ").append(method.getLoops()).append(")\n");

            if (profilers.size() > 0) {
                String methodID = method.getDefinition();
                int execCount = 0;
                BigInteger execTime = BigInteger.ZERO;
                for (ProfilerLogParser profiler : profilers) {
                    ProfilerMethodStatistics methodStats = profiler.getClassStats(classID).getMethodStats(methodID);
                    execCount += methodStats.getMethodCalls();
                    execTime = execTime.add(methodStats.getExecutionTime());
                }
                if (execCount > 0) {
                    BigInteger avg = execTime.divide(BigInteger.valueOf(execCount));
                    sb.append("\t\t    Average Execution Time: ")
                            .append(ProfilerLogParser.nanoToMilli(avg)).append(" ms\n");
                }
            }
        }

        return sb.toString();
    }

    /**
     * Returns the name of this class.
     */
    @NotNull
    public String getName() {
        return className;
    }

    /**
     * Returns the name of the package that this class belongs to.
     */
    @NotNull
    public String getPackage() {
        return fileAnalyzer.getPackage();
    }

    /**
     * Returns the ClassAnalyzer of the class that this class extends, if it is in the classpath.
     *
     * @return the associated ClassAnalyzer, or null if this class does not extend anything or if the extended class was
     *         not found.
     */
    @Nullable
    public ClassAnalyzer getExtends() {
        if (classExtends != null) {
            return fileAnalyzer.findClass(classExtends);
        }
        return null;
    }

    /**
     * Returns the ClassAnalyzers of the classes that this class implements, if they are in the classpath.
     *
     * @return the associated ClassAnalyzers, or an empty array if this class does not implement anything or if the
     *         implemented classes were not found.
     */
    public ClassAnalyzer[] getImplements() {
        List<ClassAnalyzer> result = new ArrayList<ClassAnalyzer>();
        for (String s : classImplements) {
            ClassAnalyzer ca = fileAnalyzer.findClass(s);
            if (ca != null) {
                result.add(ca);
            }
        }
        return result.toArray(new ClassAnalyzer[result.size()]);
    }

    /**
     * Returns the "weighted methods per class" of this class. The weighted methods per class is the sum of the
     * complexities of the methods in the class. Here the complexity of a method is defined as the total number of
     * conditional structures and loop structures in the method.
     */
    public int getWMC() {
        int sum = 0;
        for (MethodAnalyzer method : classMethods) {
            sum += method.getComplexity();
        }
        return sum;
    }

    /**
     * Returns the "depth of inheritance tree" of this class. The depth of inheritance tree is the number of
     * superclasses of the class. Here only the depth of the inheritance tree seen from the source files via extends
     * statements is concidered. If a class extends a library class provided by the language, the class will be treated
     * as the root of the inheritance tree. Also the implements statement are considered as inheritance (in which case
     * the largest depth of the inheritance tree is selected as the DIT of the class).
     */
    public int getDIT() {
        ClassAnalyzer ca;
        int max = 0;
        if (classExtends != null) {
            ca = fileAnalyzer.findClass(classExtends);
            if (ca != null) {
                max = Math.max(max, ca.getDIT() + 1);
            }
        }
        for (String s : classImplements) {
            ca = fileAnalyzer.findClass(s);
            if (ca != null) {
                max = Math.max(max, ca.getDIT() + 1);
            }
        }
        return max;
    }

    /**
     * Returns the "number of children" of this class. The number of children is the number of subclasses directly
     * extending or implementing the class.
     */
    public int getNOC() {
        ClassAnalyzer[] children = fileAnalyzer.getAnalyzerBatch().findChildClasses(this);
        return children.length;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj != null && obj instanceof ClassAnalyzer) {
            ClassAnalyzer other = (ClassAnalyzer) obj;
            return this.getPackage().equals(other.getPackage()) && this.getName().equals(other.getName());
        }
        return false;
    }

    /**
     * Returns code replacements that are needed for inserting the profiler code to this class.
     */
    @NotNull
    public List<StringReplacement> getProfilerCodeReplacements() {
        List<StringReplacement> replacements = new ArrayList<StringReplacement>();
        String classID = getPackage() + "." + getName();
        String content = fileAnalyzer.getContent();

        // find the beginning of this class in the file
        // HACK: assumes that there is exactly one space between "class" and the name of the class
        int pos = 0;
        while (pos >= 0) {
            pos = content.indexOf("class " + getName(), pos);
            if (pos < 0) {
                if (Tools.isDebug()) System.err.println("Class not found (1): " + getName());
                return replacements;
            }

            // check that we are not inside a comment
            String s = content.substring(0, pos);
            if (s.lastIndexOf("/*") > s.lastIndexOf("*/")) {
                continue;
            } else if (s.lastIndexOf("//") > s.lastIndexOf("\n")) {
                continue;
            }

            // found the beginning of the class's object block
            pos = content.indexOf("{", pos);
            break;
        }

        if (pos < 0) {
            if (Tools.isDebug()) System.err.println("Class not found (2): " + getName());
            return replacements;
        }
        int classStart = pos + 1;

        // find the beginning of each method
        // HACK: assumes the method definition to be in one line, except for the parameters
        // HACK: assumes the method to appear in the same order, in which case it won't be necessary to check their parameters
        for (MethodAnalyzer method : classMethods) {

            // find the beginning of the definition
            String def = method.getDefinition();
            def = def.substring(0, def.indexOf("("));

            Scanner in = new Scanner(content);

            pos = content.indexOf(def, pos);
            if (pos < 0) {
                // HACK: will not find methods if they are not formatted exactly right, or they use generics (getDefinition() will ignore generics)
                if (Tools.isDebug()) {
                    System.err.println("Method of class " + getName() + " not found: " + method.getDefinition());
                }
                continue;
            }

            // find the beginning of the method block, ignore abstract methods
            if (content.indexOf("{", pos) > content.indexOf(";", pos)) {
                continue;
            }
            pos = content.indexOf("{", pos);
            if (pos < 0) {
                continue;
            }
            int methodStart = pos + 1;

            // mark the beginning of the method
            // HACK: in case of constructors a super() or this() is hard to parse, so let's ignore those constructors
            Scanner scan = new Scanner(content.substring(methodStart));
            String s = scan.next();
            if (s.startsWith("this(") || s.startsWith("super(")) {
                if (Tools.isDebug()) {
                    System.err.println("Skipped annoying constructor: " + method.getDefinition());
                }
                continue;
            }
            replacements.add(new StringReplacement(methodStart, getMethodEnterCmd(classID, method.getDefinition())));

            // find the end of the method by counting the { and } characters
            // HACK: does not take into consideration comments and strings
            int methodEnd = -1;
            Stack<Object> stack = new Stack<Object>();
            for (int i = methodStart; i < content.length(); i++) {
                char c = content.charAt(i);
                if (c == '{') {
                    stack.push(null);
                } else if (c == '}') {
                    if (stack.size() == 0) {
                        methodEnd = i;
                        break;
                    } else {
                        stack.pop();
                    }
                }
            }
            if (methodEnd < 0) {
                if (Tools.isDebug()) {
                    System.err.println("The end of the method of class " + getName() + " not found: "
                            + method.getDefinition());
                }
                continue;
            }
            String methodContent = content.substring(0, methodEnd);

//            System.out.println(method.getDefinition());
//            System.out.println(methodStart + " - " + methodEnd);

            // find the catch blocks and return statements of the method
            // HACK: does not take strings and comments into account
            int endOfLastReturn = 0;
            for (int i = methodStart; i < methodContent.length(); i++) {

                if (methodContent.startsWith("return", i)
                        && methodContent.substring(i - 1, i + "return".length() + 1).matches("[\\s;]return[\\s;]")) {

                    // found the position of a return statement
                    int j = methodContent.indexOf(";", i) + 1;
                    replacements.add(new StringReplacement(i, j,
                            getMethodReturnCmd(classID, method.getDefinition(), methodContent.substring(i, j))));
                    endOfLastReturn = j;

                } else if (methodContent.startsWith("catch", i)
                        && methodContent.substring(i - 1, i + "catch".length() + 1).matches("[\\s\\}]catch[\\s\\(]")) {

                    // found the position of a catch statement
                    int j = methodContent.indexOf("{", i) + 1;
                    replacements.add(new StringReplacement(j, getMethodCatchCmd(classID, method.getDefinition())));
                }
            }

            // if there was no return statement at the end of the method, add the logging code for method return
            if (!methodContent.substring(endOfLastReturn).matches("[\\s\\}]*")) {
                replacements.add(new StringReplacement(methodEnd,
                        getMethodReturnCmd(classID, method.getDefinition(), "return;")));
            }
        }

        return replacements;
    }

    private static String getMethodEnterCmd(String classID, String definition) {
        return "int $$$PROFILER_RECLEVEL$$$ = net.orfjackal.sorsastus.ProfilerLogger.logEnter(\""
                + classID + "\", \"" + definition + "\")\u003B";
    }

    private static String getMethodCatchCmd(String classID, String definition) {
        return "net.orfjackal.sorsastus.ProfilerLogger.logCatch($$$PROFILER_RECLEVEL$$$, \""
                + classID + "\", \"" + definition + "\")\u003B";
    }

    private static String getMethodReturnCmd(String classID, String definition, String returnStatement) {
        return "try { " + returnStatement
                + " } finally { net.orfjackal.sorsastus.ProfilerLogger.logReturn($$$PROFILER_RECLEVEL$$$, \""
                + classID + "\", \"" + definition + "\")\u003B }";
    }
}
