// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import antlr.collections.AST;
import static net.orfjackal.sorsastus.FileAnalyzer.CursorPosition.*;
import net.orfjackal.util.*;
import org.jetbrains.annotations.*;
import util.java15grammar.*;

import java.io.*;
import java.util.*;

/**
 * Reads a Java source code file and collects metrics from it.
 *
 * @author Esko Luontola
 */
public class FileAnalyzer {

    /**
     * The batch of source code files that this file belongs to.
     */
    @NotNull
    private AnalyzerBatch analyzerBatch;

    /**
     * The file that is being analyzed.
     */
    @NotNull
    private File file;

    /**
     * The text content of the file.
     */
    @NotNull
    private String content;

    /**
     * Parsed AST tree of the source file. Is null, until the tree is created on request by calling getTree().
     */
    @Nullable
    private AST tree;

    /**
     * Number of lines in the file that contain code.
     */
    private int codeLines;

    /**
     * Number of lines in the file that do not containt code but contain comments.
     */
    private int commentLines;

    /**
     * Number of lines in the file that contain no code and no comments.
     */
    private int emptyLines;

    /* Analyzed data from this file. Initialized by analyzeStructure(), may be accessed only through Getters */
    private boolean structureAnalyzed;
    @Nullable
    private String filePackage;
    @Nullable
    private String[] fileImports;
    @Nullable
    private ClassAnalyzer[] fileClasses;
    @Nullable
    private ClassAnalyzer[] fileAnonymClasses;

    /**
     * Creates a new file analyzer and collects measures from the given file. Will read contents of the file to memory.
     *
     * @param file the file to be analyzed.
     * @throws IOException          if the file can not be read.
     * @throws NullPointerException if file is null.
     */
    public FileAnalyzer(AnalyzerBatch analyzerBatch, File file) throws IOException {
        if (analyzerBatch == null) {
            throw new NullPointerException();
        } else if (file == null) {
            throw new NullPointerException();
        }
        this.analyzerBatch = analyzerBatch;
        this.file = file;
        this.content = FileTools.readFile(file);
        this.structureAnalyzed = false;
        countLines();
    }

    /**
     * Counts the code, comment and empty lines from the content of the analyzed file. Assumes that the content of the
     * file has already been read to the <code>this.content</code> field. The results will be saved to
     * <code>this.codeLines</code>, <code>this.commentLines</code> and <code>this.emptyLines</code>.
     */
    private void countLines() {
        this.codeLines = 0;
        this.commentLines = 0;
        this.emptyLines = 0;

        CursorPosition currentPos = CODE_BLOCK;
        boolean lineContainsCode = false;
        boolean lineContainsComments = false;

        for (int i = 0; i <= content.length(); i++) {
            char c;
            if (i < content.length()) {
                c = content.charAt(i);
            } else {
                c = '\n';                                   // treat End Of File the same way as End Of Line
            }

            // locate the end of line: '\n' or EOF
            if (c == '\n') {
                switch (currentPos) {
                case CODE_BLOCK:
                    if (lineContainsCode) {                 // this line contained some code, maybe also comments
                        codeLines++;
                    } else if (lineContainsComments) {      // comment block ended on this line
                        commentLines++;
                    } else {                                // no code, no comments -> empty line
                        emptyLines++;
                    }
                    lineContainsComments = false;
                    break;

                case COMMENT_BLOCK:
                    if (lineContainsCode) {
                        codeLines++;
                    } else {
                        commentLines++;
                    }
                    lineContainsComments = true;            // comment block continues to the next line
                    break;

                case COMMENT_LINE:
                    if (lineContainsCode) {
                        codeLines++;
                    } else {
                        commentLines++;
                    }
                    currentPos = CODE_BLOCK;                // next line will be again a normal code line
                    lineContainsComments = false;
                    break;

                default:
                    assert false;
                }

                // begin analyzing the next line
                lineContainsCode = false;
                continue;
            }

            // determine where comments begin and end
            switch (currentPos) {
            case CODE_BLOCK:

                // check if here is the beginning of a comment: "/*" or "//"
                if (c == '/') {
                    char c2 = content.charAt(i + 1);
                    if (c2 == '*') {                        // beginning of a comment block
                        currentPos = COMMENT_BLOCK;
                        lineContainsComments = true;
                        i++;                                // skip analyzing the '*' character
                        continue;
                    } else if (c2 == '/') {                 // beginning of a comment line
                        currentPos = COMMENT_LINE;
                        lineContainsComments = true;
                        i++;                                // skip analyzing the '/' character
                        continue;
                    }
                }

                // check if this line contains some code, so that it is not an empty line
                if (!Character.isWhitespace(c)) {
                    lineContainsCode = true;
                }
                break;

            case COMMENT_BLOCK:

                // check if here is the end of the comment block: "*/"
                if (c == '*' && content.charAt(i + 1) == '/') {
                    currentPos = CODE_BLOCK;
                    i++;                                    // skip analyzing the '/' character
                    continue;
                }
                break;

            case COMMENT_LINE:
                // NOOP: just wait for a new line to begin
                break;

            default:
                assert false;
            }
        }
    }

    /**
     * Returns the AST tree of the source file. The tree will be created on the first call of this method. After that
     * will be returned just a reference to the tree that was created previously.
     *
     * @return the AST tree of the source file.
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    @NotNull
    private AST getTree() throws ANTLRException {
        if (tree == null) {

            // Create a scanner that reads from the input stream passed to us
            JavaLexer lexer = new JavaLexer(new StringReader(content));
            lexer.setFilename(file.getName());

            // Create a parser that reads from the scanner
            JavaRecognizer parser = new JavaRecognizer(lexer);
            parser.setFilename(file.getName());

            // start parsing at the compilationUnit rule
            parser.compilationUnit();
            tree = parser.getAST();
            assert tree != null;
        }
        return tree;
    }

    /**
     * Analyzes the AST tree of this class and saves the results for later use. Will perform the analyzing only one the
     * first run; after that the method call returns immediately.
     *
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    private void analyzeStructure() throws ANTLRException {
        if (structureAnalyzed) {
            return;
        }
        AST root = getTree();

        List<AST> packageNodes = new ArrayList<AST>();
        ASTTools.findNodes(root, JavaTreeParserTokenTypes.PACKAGE_DEF, packageNodes);

        List<AST> importNodes = new ArrayList<AST>();
        ASTTools.findNodes(root, JavaTreeParserTokenTypes.IMPORT, importNodes);

        List<AST> classNodes = new ArrayList<AST>();
        ASTTools.findNodes(root, JavaTreeParserTokenTypes.CLASS_DEF, classNodes);
        ASTTools.findNodes(root, JavaTreeParserTokenTypes.INTERFACE_DEF, classNodes);

        List<AST> anonymClassNodes = new ArrayList<AST>();
        ASTTools.findNodes(root, JavaTreeParserTokenTypes.OBJBLOCK, anonymClassNodes);

        // package of this file
        if (packageNodes.size() > 0) {
            this.filePackage = ASTTools.nodeToString(packageNodes.get(0).getFirstChild().getNextSibling());
        } else {
            this.filePackage = "";
        }

        // imports of this file
        this.fileImports = new String[importNodes.size()];
        for (int i = 0; i < importNodes.size(); i++) {
            fileImports[i] = ASTTools.nodeToString(importNodes.get(i).getFirstChild());
        }

        // classes of this file
        if (classNodes.size() + anonymClassNodes.size() > 0) {
            List<ClassAnalyzer> list = new ArrayList<ClassAnalyzer>();
            for (AST node : classNodes) {
                list.add(new ClassAnalyzer(this, node));
            }
            this.fileClasses = list.toArray(new ClassAnalyzer[list.size()]);

            // TODO: is a report of anonymous classes also needed?
            list.clear();
//            for (AST node : anonymClassNodes) {
//                list.add(new ClassAnalyzer(this, node)); // TODO: ClassAnalyzer will not work with this type of node
//            }
            this.fileAnonymClasses = list.toArray(new ClassAnalyzer[list.size()]);
        }

        // structure successfully analyzed
        structureAnalyzed = true;
    }

    /**
     * @return
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    @NotNull
    private String getFilePackage() throws ANTLRException {
        analyzeStructure();
        assert filePackage != null;
        return filePackage;
    }

    /**
     * @return
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    @NotNull
    private String[] getFileImports() throws ANTLRException {
        analyzeStructure();
        assert fileImports != null;
        return fileImports;
    }

    /**
     * @return
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    @NotNull
    private ClassAnalyzer[] getFileClasses() throws ANTLRException {
        analyzeStructure();
        assert fileClasses != null;
        return fileClasses;
    }

    /**
     * @return
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    @NotNull
    private ClassAnalyzer[] getFileAnonymClasses() throws ANTLRException {
        analyzeStructure();
        assert fileAnonymClasses != null;
        return fileAnonymClasses;
    }

    /**
     * Returns a report which contains a structural analysis of this file. The report will contain <ul><li>the name of
     * the package the file belongs to,</li><li>the number and names of packages imported,</li><li>the number and names
     * of classes in the file,</li></ul> <li>for each class:</li> <ul><li>the number and names of the methods in the
     * class,</li></ul> <li>for each method:</li> <ul><li>the number of if and switch statements,</li><li>the number of
     * for, while and do-while statements.</li></ul></ul> The report will end with the line freed character '\n'.
     *
     * @throws ANTLRException if there was a parse exception in parsing the source file.
     */
    @NotNull
    public String getStructureReport() throws ANTLRException {
        StringBuilder sb = new StringBuilder();

        // package report
        String filePackage = getFilePackage();
        if (filePackage.length() > 0) {
            sb.append("    Package:\n");
            sb.append('\t').append(filePackage).append('\n');
        }

        // imports report
        String[] fileImports = getFileImports();
        if (fileImports.length > 0) {
            sb.append("    Imports:\n");
            for (String s : fileImports) {
                sb.append('\t').append(s).append('\n');
            }
        }

        // classes report
        ClassAnalyzer[] fileClasses = getFileClasses();
        ClassAnalyzer[] fileAnonymClasses = getFileAnonymClasses();
        if (fileClasses.length > 0) {
            sb.append("    Classes:\n");
            for (ClassAnalyzer c : fileClasses) {
                sb.append(c.getStructureReport());
            }
            for (ClassAnalyzer c : fileAnonymClasses) {
                sb.append(c.getStructureReport());
            }
        }
        return sb.toString();
    }

    public AnalyzerBatch getAnalyzerBatch() {
        return analyzerBatch;
    }

    /**
     * Returns the file that was analyzed.
     */
    @NotNull
    public File getFile() {
        return file;
    }

    /**
     * Returns the total number of lines in the file, including all comments and empty lines.
     */
    public int getTotalLines() {
        return codeLines + commentLines + emptyLines;
    }

    /**
     * Returns the number of code lines in the file, excluding comment lines and empty lines.
     */
    public int getCodeLines() {
        return codeLines;
    }

    /**
     * Returns the number of lines in the file that contain only comments and no code.
     */
    public int getCommentLines() {
        return commentLines;
    }

    /**
     * Returns the number of lines that contain no code and no comments.
     */
    public int getEmptyLines() {
        return emptyLines;
    }

    /**
     * Returns the name of the package that this file belongs to.
     */
    @NotNull
    public String getPackage() {
        try {
            return getFilePackage();
        } catch (ANTLRException e) {
            if (Tools.isDebug()) e.printStackTrace();
            return "";
        }
    }

    /**
     * Returns the classes that are part of this FileAnalyzer.
     */
    @NotNull
    public ClassAnalyzer[] getClasses() {
        try {
            ClassAnalyzer[] result = new ClassAnalyzer[getFileClasses().length];
            System.arraycopy(getFileClasses(), 0, result, 0, result.length);
            return result;
        } catch (ANTLRException e) {
            if (Tools.isDebug()) e.printStackTrace();
            return new ClassAnalyzer[0];
        }
    }

    /**
     * Returns the ClassAnalyzer of the specified if it is contained in this FileAnalyzer. Will not look into other
     * files for the class.
     *
     * @param classPackage the package of the class to be looked for.
     * @param className    the name of the class to be looked for.
     * @return the ClassAnalyzer of the class if it has the same package and class name as a class in this FileAnalyzer.
     *         Otherwise will return null.
     */
    @Nullable
    public ClassAnalyzer getClass(String classPackage, String className) {
        try {
            if (getFilePackage().equals(classPackage)) {             // found the right package
                ClassAnalyzer[] classes = getFileClasses();
                for (ClassAnalyzer ca : classes) {
                    if (ca.getName().equals(className)) {           // found the right class
                        return ca;
                    }
                }
            }
        } catch (ANTLRException e) {
            if (Tools.isDebug()) e.printStackTrace();
        }
        return null;
    }

    /**
     * Tries to find a class of the specified name from the list of classes that have been imported to this file.
     * Returns the ClassAnalyzer associated with that class, or null if such a class was not found in the classpath.
     *
     * @return the ClassAnalyzer of the class, or null if it was not found.
     * @throws NullPointerException if any parameter is null.
     */
    @Nullable
    public ClassAnalyzer findClass(String className) {
        if (className == null) {
            throw new NullPointerException();
        }

        // initialize this FileAnalyzer, if the analyzing has not yet been done
        String filePackage;
        String[] fileImports;
        ClassAnalyzer[] fileClasses;
        try {
            filePackage = getFilePackage();
            fileImports = getFileImports();
            fileClasses = getFileClasses();
        } catch (ANTLRException e) {
            if (Tools.isDebug()) e.printStackTrace();
            return null;
        }

        // check for an absolute class name (such as java.lang.Object)
        if (className.indexOf('.') >= 0) {
            String s1 = className.substring(0, className.lastIndexOf('.'));
            String s2 = className.substring(s1.length() + 1);
            return analyzerBatch.findClass(s1, s2);
        }

        // go through the import statements in order:

        // 1st priority: Types in the same compilation unit
        for (ClassAnalyzer ca : fileClasses) {
            if (ca.getName().equals(className)) {
                return ca; // TODO: not tested
            }
        }

        // 2nd priority: Single-Type-Import (§7.5.1)
        ClassAnalyzer ca;
        for (String s : fileImports) {
            if (!s.regionMatches(s.length() - 2, ".*", 0, 2)) {
                String importPackage = s.substring(0, s.lastIndexOf('.'));
                String importClass = s.substring(importPackage.length() + 1);
                if (!importClass.equals(className)) {
                    continue;
                }
                ca = analyzerBatch.findClass(importPackage, className);
                if (ca != null) {
                    return ca;
                }
            }
        }

        // 3rd priority: Types in the same package
        ca = analyzerBatch.findClass(filePackage, className);
        if (ca != null) {
            return ca;  // TODO: not tested
        }

        // 4th priority: Type-Import-on-Demand (§7.5.2)
        for (String s : fileImports) {
            if (s.regionMatches(s.length() - 2, ".*", 0, 2)) {
                String importPackage = s.substring(0, s.length() - 2);
                ca = analyzerBatch.findClass(importPackage, className);
                if (ca != null) {
                    return ca;
                }
            }
        }

        // class not found
        return null;
    }

    /**
     * Returns the code of this source code file with the profiler code inserted to it. Does not modify the file
     * itself.
     */
    @NotNull
    public String getProfilerCode() throws ANTLRException {
        analyzeStructure();
        List<StringReplacement> replacements = new ArrayList<StringReplacement>();

        for (ClassAnalyzer classAnalyzer : fileClasses) {
            replacements.addAll(classAnalyzer.getProfilerCodeReplacements());
        }
        for (ClassAnalyzer classAnalyzer : fileAnonymClasses) {
            replacements.addAll(classAnalyzer.getProfilerCodeReplacements());
        }
        return StringReplacement.replace(content, replacements);
    }

    public String getContent() {
        return content;
    }

    protected enum CursorPosition {
        CODE_BLOCK, COMMENT_BLOCK, COMMENT_LINE
    }

    public List<Object> getList() {
        return null;
    }
}
