// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import antlr.collections.AST;
import net.orfjackal.util.ASTTools;
import org.jetbrains.annotations.*;
import util.java15grammar.JavaTreeParserTokenTypes;

/**
 * Collects metrics from the structure of a method.
 *
 * @author Esko Luontola
 */
public class MethodAnalyzer {

    @NotNull
    private ClassAnalyzer classAnalyzer;

    @NotNull
    private AST methodNode;

    /* Analyzed data from this method. Initialized by analyzeStructure() */
    @NotNull
    private String methodDefinition;
    private int methodIfCount;
    private int methodSwitchCount;
    private int methodForCount;
    private int methodWhileCount;
    private int methodDoWhileCount;

    /**
     * Creates a new Method and analyzes the structure of the method.
     *
     * @param classAnalyzer the class that this method belongs to.
     * @param methodNode    structure of the method. The type of the node must be CTOR_DEF or METHOD_DEF.
     * @throws NullPointerException     if any parameter is null.
     * @throws IllegalArgumentException if methodNode is of wrong type.
     */
    public MethodAnalyzer(ClassAnalyzer classAnalyzer, AST methodNode) {
        if (classAnalyzer == null) {
            throw new NullPointerException();
        } else if (methodNode == null) {
            throw new NullPointerException();
        } else if (methodNode.getType() != JavaTreeParserTokenTypes.CTOR_DEF
                && methodNode.getType() != JavaTreeParserTokenTypes.METHOD_DEF) {
            throw new IllegalArgumentException("methodNode.getType() = " + methodNode.getType());
        }
        this.classAnalyzer = classAnalyzer;
        this.methodNode = methodNode;
        analyzeStructure();
    }

    /**
     * Analyzes the AST tree of this method and saves the results for later use.
     */
    private void analyzeStructure() {

        // components of the method definition
        @NotNull AST mModifiers = null;
        @Nullable AST mType = null;         // null for constructors
        @NotNull AST mIdent = null;
        @NotNull AST mParameters = null;
        @Nullable AST mSlist = null;        // null if abstract method
        @Nullable AST mThrows = null;       // null if does not throw anything

        // find the components of the method definition
        AST m = methodNode.getFirstChild();
        while (m != null) {
            switch (m.getType()) {
            case JavaTreeParserTokenTypes.MODIFIERS:
                mModifiers = m;
                break;
            case JavaTreeParserTokenTypes.TYPE:
                mType = m;
                break;
            case JavaTreeParserTokenTypes.IDENT:
                mIdent = m;
                break;
            case JavaTreeParserTokenTypes.PARAMETERS:
                mParameters = m;
                break;
            case JavaTreeParserTokenTypes.SLIST:
                mSlist = m;
                break;
            case JavaTreeParserTokenTypes.LITERAL_throws:
                mThrows = m;
                break;
            default:
                System.err.println(m.toStringTree());
                assert false : "m.getType() = " + m.getType();
                break;
            }
            m = m.getNextSibling();
        }
        assert mModifiers != null;
        assert mIdent != null;
        assert mParameters != null;

        // begin building the method definition
        StringBuilder sb = new StringBuilder();

        // get the modifiers
        m = mModifiers.getFirstChild();
        while (m != null) {
            if (m.getType() != JavaTreeParserTokenTypes.ANNOTATION) {
                sb.append(m.getText()).append(' ');
            }
            m = m.getNextSibling();
        }

        // get the type
        if (mType != null) {
            if (mType.getFirstChild().getType() == JavaTreeParserTokenTypes.ARRAY_DECLARATOR) {
                sb.append(mType.getFirstChild().getFirstChild().getText()).append("[]");
            } else {
                sb.append(mType.getFirstChild().getText());
            }
            sb.append(' ');
        }

        // get the method identifier
        sb.append(mIdent.getText()).append('(');
        m = mParameters.getFirstChild();
        while (m != null) {

            // get the method parameters
            AST child = m.getFirstChild();                  // MODIFIERS
            AST pType = child = child.getNextSibling();     // TYPE
            AST pIdent = child.getNextSibling();            // parameter name

            if (pType.getFirstChild().getType() == JavaTreeParserTokenTypes.ARRAY_DECLARATOR) {
                sb.append(pType.getFirstChild().getFirstChild().getText()).append("[]");
            } else {
                sb.append(pType.getFirstChild().getText());
            }
            sb.append(' ');
            sb.append(pIdent.getText());

            m = m.getNextSibling();
            if (m != null) {
                sb.append(", ");
            }
        }
        sb.append(')');

        // end of building the method definition
        this.methodDefinition = sb.toString();

        // get the complexity of the method
        this.methodIfCount = ASTTools.getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_if,
                JavaTreeParserTokenTypes.OBJBLOCK);         // do not advance into anonymous inner classes
        this.methodSwitchCount = ASTTools.getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_switch,
                JavaTreeParserTokenTypes.OBJBLOCK);
        this.methodForCount = ASTTools.getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_for,
                JavaTreeParserTokenTypes.OBJBLOCK);
        this.methodWhileCount = ASTTools.getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_while,
                JavaTreeParserTokenTypes.OBJBLOCK);         // does NOT count the do-while structures
        this.methodDoWhileCount = ASTTools.getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_do,
                JavaTreeParserTokenTypes.OBJBLOCK);
    }

    @NotNull
    public String getDefinition() {
        return methodDefinition;
    }

    public int getConditionals() {
        return methodIfCount + methodSwitchCount;
    }

    public int getLoops() {
        return methodForCount + methodWhileCount + methodDoWhileCount;
    }

    public int getComplexity() {
        return getConditionals() + getLoops();
    }
}
