// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import org.jetbrains.annotations.NotNull;

import java.io.*;
import java.math.*;
import java.text.*;
import java.util.*;

/**
 * Reads the log generated by a runtime profiler. The specifications of the log file are as follows:
 * <p/>
 * Each thread has an unique log file.<br />All log files are in one directory and their file name will end with
 * ".log".<br />The log files are UTF-8 text files.
 * <p/>
 * The format of the rows in the log file is as follows:
 * <p/>
 * <tt>TIMESTAMP<b>;</b>OPCODE<b>;</b>RECLEVEL<b>;</b>CLASS<b>;</b>METHOD</tt>
 * <p/>
 * where
 * <p/>
 * <tt>TIMESTAMP</tt><br />Execution time since the beginning of the program in nanoseconds, excluding the time used by
 * the profiler itself. The format of TIMESTAMP is an integer number of unlimited magnitude (BigInteger).
 * <p/>
 * <tt>OPCODE</tt><br />A string indicating whether this log entry denotes when the method was entered, an exception was
 * caught or there execution of the method ended (with a return statement or end of code block). Possible values are
 * "ENTER", "CATCH" and "RETURN".
 * <p/>
 * <tt>RECLEVEL</tt><br />Recursion level; the number of method calls in the stack before the current method call. The
 * main method of the program or the beginning of a new thread will use the number 0. If the number of method A is N and
 * method A calls method B, then the number of method B is N+1. The format of RECLEVEL is an integer number (max
 * 32bit).
 * <p/>
 * <tt>CLASS</tt><br />The name of the class where the method resides. Must be an unique identifier for each class
 * (package name and class name). For example "java.lang.String". Ignored when OPCODE != "ENTER".
 * <p/>
 * <tt>METHOD</tt><br />The name of the method. Must be an unique identifier for each method in the class. For example
 * "compareTo(Object)". Ignored when OPCODE != "ENTER".
 * <p/>
 * All method calls in the log file should have ended at the end of the log file. If that is not possible, the following
 * lines can be added to the end of the log file:<br /><tt>TIMESTAMP<b>;CATCH;0;FORCED_EXIT;FORCED_EXIT</b><br
 * />TIMESTAMP<b>;RETURN;0;FORCED_EXIT;FORCED_EXIT</b></tt>
 *
 * @author Esko Luontola
 */
public class ProfilerLogParser {

    public static final String LOG_FILE_TYPE = "log";
    public static final String LOG_FORCED_EXIT = "FORCED_EXIT";
    public static final String LOG_ENTRY_DELIMITER = ";";

    public static final int LOG_ENTRIES = 5;
    public static final int LOG_ENTRY_TIMESTAMP = 0;
    public static final int LOG_ENTRY_OPCODE = 1;
    public static final int LOG_ENTRY_RECLEVEL = 2;
    public static final int LOG_ENTRY_CLASS = 3;
    public static final int LOG_ENTRY_METHOD = 4;

    public static final int REPORT_MILLISECOND_FRACTIONS = 3;

    public enum OpCode {
        ENTER, CATCH, RETURN
    }

    @NotNull
    private File[] logFiles;

    @NotNull
    private Map<String, ProfilerClassStatistics> classes;

    /**
     * Reads the profiler log files and generates a report of the execution times.
     *
     * @param logFiles the log files to be analyzed.
     * @throws IOException          if one of the parameters is not a file or it can not be read.
     * @throws NullPointerException if logFiles or one of its values is null.
     */
    public ProfilerLogParser(File[] logFiles) throws IOException {
        if (logFiles == null) {
            throw new NullPointerException("logFiles is null");
        }
        this.logFiles = new File[logFiles.length];
        this.classes = new HashMap<String, ProfilerClassStatistics>();
        for (int i = 0; i < logFiles.length; i++) {
            File logFile = logFiles[i];
            if (logFile == null) {
                throw new NullPointerException("logFiles[" + i + "] is null");
            } else if (!logFile.isFile()) {
                throw new IOException("Not a file: " + logFile);
            } else if (!logFile.canRead()) {
                throw new IOException("Unable to read file: " + logFile);
            }
            this.logFiles[i] = logFile;
        }
        analyze();
    }

    public ProfilerClassStatistics getClassStats(String classID) {
        ProfilerClassStatistics classStats = classes.get(classID);
        if (classStats == null) {
            classStats = new ProfilerClassStatistics(classID);
            classes.put(classID, classStats);
        }
        return classStats;
    }

    /**
     * Analyzes all the files in the logFiles array.
     */
    private void analyze() throws IOException {
        classes.clear();
        for (File file : logFiles) {
            Scanner in = new Scanner(file);
            try {
                analyze(in);
            } catch (ParseException e) {
                throw new IOException("Parse exception in file: " + file + "\n" + e.getMessage());
            }
        }
    }

    /**
     * Analyzes one log file.
     *
     * @param in an unused Scanner ready for reading the file.
     * @throws ParseException if one of the rows in the log file was in wrong format. The message of the exception will
     *                        contain an explanation.
     */
    private void analyze(Scanner in) throws ParseException {

        // stack of all method calls in the thread
        // NOTICE: the size of the stack equals the reclevel of the next new method call
        Stack<ProfilerMethodCall> stack = new Stack<ProfilerMethodCall>();

        while (in.hasNextLine()) {
            @NotNull String line = in.nextLine();
            @NotNull BigInteger timestamp;
            @NotNull OpCode opcode;
            int reclevel;
            @NotNull String classID;
            @NotNull String methodID;

            // parse the entries in the line
            String[] entries = line.split(LOG_ENTRY_DELIMITER);
            if (entries.length != LOG_ENTRIES) {
                throw new ParseException(LOG_ENTRIES + " entries excepted, "
                        + entries.length + " found on line: " + line, 0);
            }
            try {
                timestamp = new BigInteger(entries[LOG_ENTRY_TIMESTAMP]);
            } catch (NumberFormatException e) {
                throw new ParseException("Not a valid TIMESTAMP on line: " + line, 0);
            }
            if (timestamp.compareTo(BigInteger.ZERO) < 0) {
                throw new ParseException("Negative TIMESTAMP value " + timestamp + " on line: " + line, 0);
            }
            try {
                opcode = OpCode.valueOf(entries[LOG_ENTRY_OPCODE]);
            } catch (IllegalArgumentException e) {
                throw new ParseException("Not a valid OPCODE on line: " + line, 0);
            }
            try {
                reclevel = Integer.parseInt(entries[LOG_ENTRY_RECLEVEL]);
            } catch (NumberFormatException e) {
                throw new ParseException("Not a valid RECLEVEL on line: " + line, 0);
            }
            if (reclevel < 0) {
                throw new ParseException("Negative RECLEVEL value " + reclevel + " on line: " + line, 0);
            }
            classID = entries[LOG_ENTRY_CLASS];
            methodID = entries[LOG_ENTRY_METHOD];

            // start or stop the method call timers when the method is entered or exited
            switch (opcode) {
            case ENTER:
                if (stack.size() != reclevel) {
                    throw new ParseException("Illegal RECLEVEL value ("
                            + stack.size() + " expected, "
                            + reclevel + " found) on line: " + line, 0);
                }
                if (stack.size() > 0) {
                    stack.peek().pauseTimer(timestamp);
                }
                stack.push(new ProfilerMethodCall(this, timestamp, classID, methodID));
                break;

            case CATCH:
                if (stack.size() <= reclevel) {
                    throw new ParseException("Illegal RECLEVEL value (<"
                            + stack.size() + " expected, "
                            + reclevel + " found) on line: " + line, 0);
                }
                // must take into account that the exception could have been thrown by a class outside the profilation
                while ((stack.size() - 1) > reclevel) {
                    stack.pop().exitMethod(timestamp);
                    stack.peek().resumeTimer(timestamp);
                }
                break;

            case RETURN:
                if ((stack.size() - 1) != reclevel) {
                    throw new ParseException("Illegal RECLEVEL value ("
                            + (stack.size() - 1) + " expected, "
                            + reclevel + " found) on line: " + line, 0);
                }
                stack.pop().exitMethod(timestamp);
                if (stack.size() > 0) {
                    stack.peek().resumeTimer(timestamp);
                }
                break;

            default:
                assert false;
                break;
            }
        }

        // check that all methods were exited nicely
        if (stack.size() > 0) {
            throw new ParseException("Incomplete log file, " + stack.size() + " unfinished method calls.", 0);
        }
    }

    /**
     * Returns a runtime report based on the profiler logs.
     */
    @NotNull
    // TODO: integrate this method to the structure report
    public String getProfilerReport() {
        StringBuilder sb = new StringBuilder();

        String[] classIDs = classes.keySet().toArray(new String[0]);
        Arrays.sort(classIDs);

        for (String classID : classIDs) {
            ProfilerClassStatistics classStats = getClassStats(classID);
            String[] methodIDs = classStats.getMethods().keySet().toArray(new String[0]);
            Arrays.sort(methodIDs);

            sb.append(classID).append("\n");

            sb.append("\tTotal Execution Time: ").append(nanoToMilli(classStats.getTotalExecutionTime())).append(
                    " ms\n");

            for (String methodID : methodIDs) {
                ProfilerMethodStatistics methodStats = classStats.getMethodStats(methodID);
                sb.append("\t").append(methodID).append("\n");
                sb.append("\t    Method Calls: ").append(methodStats.getMethodCalls()).append("\n");
                sb.append("\t    Execution Time: ").append(nanoToMilli(methodStats.getExecutionTime())).append(" ms\n");
                sb.append("\t    Inclusive Execution Time: ")
                        .append(nanoToMilli(methodStats.getInclusiveExecutionTime())).append(" ms\n");
            }
            sb.append("\n");
        }
        return sb.toString();
    }

    /**
     * Converts nanoseconds to milliseconds with 3 decimal.
     */
    public static String nanoToMilli(BigInteger nanoTime) {
        BigDecimal milliTime = new BigDecimal(nanoTime);
        //milliTime = milliTime.divide(BigDecimal.valueOf(1000), RoundingMode.HALF_UP);
        //milliTime = milliTime.divide(BigDecimal.valueOf(1000));
        milliTime = milliTime.divide(BigDecimal.valueOf(1000000));
        NumberFormat nf = NumberFormat.getInstance();
        nf.setMinimumFractionDigits(REPORT_MILLISECOND_FRACTIONS);
        nf.setMaximumFractionDigits(REPORT_MILLISECOND_FRACTIONS);
        return nf.format(milliTime.doubleValue());
    }
}
