// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import org.jetbrains.annotations.NotNull;

import java.math.BigInteger;

/**
 * Represents an ongoing method call in the stack. Takes case of calculating the execution time within the method.
 * Records the information to the profiler results.
 */
public class ProfilerMethodCall {

    @NotNull
    private ProfilerLogParser profiler;

    @NotNull
    private String classID;

    @NotNull
    private String methodID;

    /**
     * When this method call was made.
     */
    @NotNull
    private BigInteger startTimestamp;

    /**
     * The last time the method was entered or exited (temporarily).
     */
    @NotNull
    private BigInteger lastTimestamp;

    /**
     * Total execution time spent in the method (excluding methods calls made from this method).
     */
    @NotNull
    private BigInteger execTime;

    /**
     * True if this method is executing right now. False if this method has called another method or this method has
     * exited.
     */
    private boolean running;

    /**
     * True if this method has exited, otherwise false.
     */
    private boolean exited;

    /**
     * Starts a method call.
     *
     * @param timestamp the timestamp of the event.
     * @param classID   identifier of the class.
     * @param methodID  identifier of the method.
     * @throws NullPointerException if any parameter is null.
     */
    public ProfilerMethodCall(ProfilerLogParser profiler, BigInteger timestamp, String classID, String methodID) {
        if (profiler == null) {
            throw new NullPointerException("profiler is null");
        } else if (timestamp == null) {
            throw new NullPointerException("timestamp is null");
        } else if (classID == null) {
            throw new NullPointerException("classID is null");
        } else if (methodID == null) {
            throw new NullPointerException("methodID is null");
        }
        this.profiler = profiler;
        this.classID = classID;
        this.methodID = methodID;
        this.startTimestamp = timestamp;
        this.lastTimestamp = timestamp;
        this.execTime = BigInteger.ZERO;
        this.running = true;
        this.exited = false;
    }

    /**
     * Pauses the execution of the method. Used when the method calls another method.
     *
     * @param timestamp the timestamp of the event.
     * @throws NullPointerException if any parameter is null.
     */
    public void pauseTimer(BigInteger timestamp) {
        if (timestamp == null) {
            throw new NullPointerException();
        }
        if (!running) {
            throw new IllegalStateException("Can not pause at " + timestamp + ", the method "
                    + classID + "." + methodID + " has not been executing since " + lastTimestamp + ".");
        }
        assert !exited;
        execTime = execTime.add(timestamp.subtract(lastTimestamp));
        lastTimestamp = timestamp;
        running = false;
    }

    /**
     * Resumes the execution of the method. Used when returning from a method call made by this method (through a clean
     * return or an exception).
     *
     * @param timestamp the timestamp of the event.
     * @throws NullPointerException if any parameter is null.
     */
    public void resumeTimer(BigInteger timestamp) {
        if (timestamp == null) {
            throw new NullPointerException();
        }
        if (running) {
            throw new IllegalStateException("Can not resume at " + timestamp + ", the method "
                    + classID + "." + methodID + " has already been executing since " + lastTimestamp + ".");
        }
        assert !exited;
        lastTimestamp = timestamp;
        running = true;
    }

    /**
     * Marks the end of the method call. Used when returning from this method call (through a clean return or an
     * exception). Records the time spent in the method.
     *
     * @param timestamp the timestamp of the event.
     * @throws NullPointerException if any parameter is null.
     */
    public void exitMethod(BigInteger timestamp) {
        if (timestamp == null) {
            throw new NullPointerException();
        }
        pauseTimer(timestamp);
        assert !exited;
        exited = true;
        profiler.getClassStats(classID).getMethodStats(methodID).registerMethodCall(this);
    }

    @NotNull
    public String getClassID() {
        return classID;
    }

    @NotNull
    public String getMethodID() {
        return methodID;
    }

    @NotNull
    public BigInteger getStartTimestamp() {
        return startTimestamp;
    }

    public boolean isExited() {
        return exited;
    }

    @NotNull
    public BigInteger getExecutionTime() {
        assert exited;
        return execTime;
    }

    @NotNull
    public BigInteger getInclusiveExecutionTime() {
        assert exited;
        return lastTimestamp.subtract(startTimestamp);
    }
}
