// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import org.jetbrains.annotations.NotNull;

import java.math.BigInteger;

/**
 * Collects the profiler log statistics of one method.
 */
public class ProfilerMethodStatistics {

    @NotNull
    private String classID;

    @NotNull
    private String methodID;

    private int methodCalls;

    @NotNull
    private BigInteger executionTime;

    @NotNull
    private BigInteger inclusiveExecutionTime;

    public ProfilerMethodStatistics(String classID, String methodID) {
        if (classID == null) {
            throw new NullPointerException("classID is null");
        } else if (methodID == null) {
            throw new NullPointerException("methodID is null");
        }
        this.classID = classID;
        this.methodID = methodID;
        this.methodCalls = 0;
        this.executionTime = BigInteger.ZERO;
        this.inclusiveExecutionTime = BigInteger.ZERO;
    }

    public void registerMethodCall(ProfilerMethodCall methodCall) {
        if (methodCall == null) {
            throw new NullPointerException("methodCall is null");
        } else if (!methodCall.isExited()) {
            throw new IllegalArgumentException("Method call of "
                    + methodCall.getClassID() + "." + methodCall.getMethodID()
                    + " at " + methodCall.getStartTimestamp() + " has not exited");
        } else if (!classID.equals(methodCall.getClassID())) {
            throw new IllegalArgumentException("classID " + classID + " expected, "
                    + methodCall.getClassID() + " found.");
        } else if (!methodID.equals(methodCall.getMethodID())) {
            throw new IllegalArgumentException("methodID " + methodID + " expected, "
                    + methodCall.getMethodID() + " found.");
        }
        methodCalls++;
        executionTime = executionTime.add(methodCall.getExecutionTime());
        inclusiveExecutionTime = inclusiveExecutionTime.add(methodCall.getInclusiveExecutionTime());
    }

    public int getMethodCalls() {
        return methodCalls;
    }

    @NotNull
    public BigInteger getExecutionTime() {
        return executionTime;
    }

    @NotNull
    public BigInteger getAvgExecutionTime() {
        if (getMethodCalls() > 0) {
            return getExecutionTime().divide(BigInteger.valueOf(getMethodCalls()));
        } else {
            return BigInteger.ZERO;
        }
    }

    @NotNull
    public BigInteger getInclusiveExecutionTime() {
        return inclusiveExecutionTime;
    }
}
