// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import net.orfjackal.util.Tools;
import net.orfjackal.util.commandline.*;

import java.io.*;
import java.text.ParseException;

/**
 * Handles the command line parameters that this program takes and executes the program.
 *
 * @author Esko Luontola
 */
public class SorsastusLauncher extends CommandLineParser {

    private static final int LOC_OPTION_CODE_LINES = 1;
    private static final int LOC_OPTION_COMMENT_LINES = 2;
    private static final int LOC_OPTION_EMPTY_LINES = 4;
    private static final int LOC_OPTION_TOTAL_LINES = 7;
    private static final int STRUCTURE_OPTION_OOMETRICS = 1;

    private int locOptions = 0;
    private int structureOptions = 0;
    private boolean doStructure = false;

    private AnalyzerBatch analyzer;

    public SorsastusLauncher(AnalyzerBatch analyzerBatch) {
        super(Sorsastus.APP_COMMAND, Sorsastus.APP_DESCRIPTION);
        if (analyzerBatch == null) {
            throw new NullPointerException("analyzerBatch is null");
        }
        this.analyzer = analyzerBatch;

        addOption(new Option("c", "code-lines",
                "include effective lines of code to the LOC count (default, if none of `-mets' defined)") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_CODE_LINES;
            }
        });
        addOption(new Option("m", "comment-lines", "include comment lines to the LOC count") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_COMMENT_LINES;
            }
        });
        addOption(new Option("e", "empty-lines", "include empty lines to the LOC count") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_EMPTY_LINES;
            }
        });
        addOption(new Option("t", "total-lines", "include all lines to the LOC count (the same as `-cme')") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_TOTAL_LINES;
            }
        });
        addOption(new Option("s", "structure", "output a detailed report of the structure of the source code files") {
            public void onSpecified(String arg) {
                doStructure = true;
            }
        });
        addOption(new Option("o", "oo-metrics", "include object-oriented metrics to the structure report (slow)") {
            public void onSpecified(String arg) {
                structureOptions = structureOptions | STRUCTURE_OPTION_OOMETRICS;
            }
        });
        // TODO: remove the following option
        addOption(new Option(null, "write-profiler-files", "DIRECTORY", true,
                "writes profiled versions of the input files to the specified directory, will overwrite all files") {
            public void onSpecified(String arg) throws ParseException {
                File dir = new File(arg);
                if (dir.exists() && dir.isFile()) {
                    throw new ParseException("Not a directory: " + dir, 0);
                }
                if (!dir.isDirectory() && !dir.mkdirs()) {
                    throw new ParseException("Unable to create directory: " + dir, 0);
                }
                analyzer.insertProfilerCode(dir);
                System.exit(0);
            }
        });
        // TODO: remove the following option, or integrate the profiler report to the structure report
        addOption(new Option(null, "read-profiler-logs", "DIRECTORY", true,
                "reads a directory of runtime profiler logs and outputs a runtime report based on them") {
            public void onSpecified(String arg) throws ParseException {
                // TODO: remove this option when this program does the compiling and running of the profiled program
                File dir = new File(arg);
                if (!dir.isDirectory()) {
                    throw new ParseException("Not a directory: " + dir, 0);
                } else if (!dir.canRead()) {
                    throw new ParseException("Unable to read directory: " + dir, 0);
                }
                File[] logFiles = dir.listFiles(new FilenameFilter() {
                    public boolean accept(File dir, String name) {
                        return name.matches(".*?\\." + ProfilerLogParser.LOG_FILE_TYPE + "\\z");
                    }
                });
                try {
                    ProfilerLogParser profiler = new ProfilerLogParser(logFiles);
                    System.out.println(profiler.getProfilerReport());
                } catch (IOException e) {
                    if (Tools.isDebug()) e.printStackTrace();
                    throw new ParseException(e.getMessage(), 0);
                }
                System.exit(0);
            }
        });
        addOption(new Option(null, "profiler-main", "CLASS", true,
                "full name of the main class that will start the profiled program; specify this to make an automated runtime analysis") {
            public void onSpecified(String arg) throws ParseException {
                analyzer.setProfilerMainClass(arg);
            }
        });
        addOption(new Option("P", "profiler-parameter", "STRING", true,
                "a runtime parameter for the profiled program, define once for each parameter (optional)") {
            public void onSpecified(String arg) throws ParseException {
                analyzer.addProfilerParameter(arg);
            }
        });
        addOption(new Option(null, "profiler-runs", "INTEGER", true,
                "how many times the profiled program will be executed (default: 1)") {
            public void onSpecified(String arg) throws ParseException {
                try {
                    analyzer.setProfilerRuns(Integer.parseInt(arg));
                } catch (NumberFormatException e) {
                    throw new ParseException("Not an integer number: " + arg, 0);
                }
            }
        });
        addOption(new Option(null, "jdk-path", "DIRECTORY", true,
                "directory of the JDK that will be used for compiling and running the profiled program (optional)") {
            public void onSpecified(String arg) throws ParseException {
                File path = new File(arg);
                if (!path.isDirectory()) {
                    throw new ParseException("No such directory: " + arg, 0);
                }
                analyzer.setProfilerJDKPath(path);
            }
        });
        addDefaultHelpOption();
        addOption(new Option(null, "version", "output version information and exit") {
            public void onSpecified(String arg) {
                System.out.println(Sorsastus.APP_NAME + " " + Sorsastus.APP_VERSION);
                System.out.println(Sorsastus.COPYRIGHT);
                System.exit(0);
            }
        });

        addArgument(new FileArgument("FILE_OR_DIR", true, true) {
            public void onSpecified(File[] files) throws ParseException {
                for (File file : files) {
                    try {
                        SorsastusLauncher.this.analyzer.addFile(file);
                    } catch (IOException e) {
                        throw new ParseException(e.getMessage(), 0);
                    }
                }
            }
        });
    }

    @Override
    public void parse(String[] params) {
        try {
            super.parse(params);
        } catch (ParseException e) {
            System.out.println(Sorsastus.APP_COMMAND + ": " + e.getMessage());
            System.out.println("Try `" + Sorsastus.APP_COMMAND + " --help' for more information.");
            System.exit(1);
        }

        // set the options
        if (locOptions == 0) {
            locOptions = locOptions | LOC_OPTION_CODE_LINES;
        }
        analyzer.setCountCodeLines((locOptions & LOC_OPTION_CODE_LINES) != 0);
        analyzer.setCountCommentLines((locOptions & LOC_OPTION_COMMENT_LINES) != 0);
        analyzer.setCountEmptyLines((locOptions & LOC_OPTION_EMPTY_LINES) != 0);
        analyzer.setIncludeOOMetrics((structureOptions & STRUCTURE_OPTION_OOMETRICS) != 0);

        // print the requested report
        if (analyzer.isProfilerEnabled()) {
            analyzer.doProfiler();
            try {
                System.out.print(analyzer.getStructureReport());
            } catch (ANTLRException e) {
                System.out.println(Sorsastus.APP_COMMAND + ": " + e.getMessage());
                System.exit(1);
            }
        } else if (doStructure) {
            try {
                System.out.print(analyzer.getStructureReport());
            } catch (ANTLRException e) {
                System.out.println(Sorsastus.APP_COMMAND + ": " + e.getMessage());
                System.exit(1);
            }
        } else {
            System.out.print(analyzer.getLineCountReport());
        }
    }
}
