// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.util;

import antlr.collections.AST;
import org.jetbrains.annotations.NotNull;

import java.util.List;

/**
 * Library class of AST tree operations.
 *
 * @author Esko Luontola
 */
public class ASTTools {

    /**
     * Finds from the provided AST tree all the nodes of the type nodeType and adds them to the result list.
     *
     * @param node     the root node of the tree from where to search. If null, will do nothing.
     * @param nodeType the type of the nodes to be searched for.
     * @param result   the nodes that were found will be added to this list.
     * @param exclude  the types of the nodes into which the algorithm should not dig into.
     * @throws NullPointerException if result is null and a matching node was found.
     */
    public static void findNodes(AST node, int nodeType, List<AST> result, int... exclude) {
        while (node != null) {
            if (node.getNumberOfChildren() > 0) {
                boolean advance = true;
                for (int ex : exclude) {
                    if (node.getType() == ex) {
                        advance = false;
                    }
                }
                if (advance) {
                    findNodes(node.getFirstChild(), nodeType, result, exclude);
                }
            }
            if (node.getType() == nodeType) {
                //System.out.println(node.toStringTree()); // DEBUG
                result.add(node);
            }
            node = node.getNextSibling();
        }
    }

    /**
     * Appends the text value of an AST node and all of its child nodes.
     *
     * @param node the node to be converted to a string. If null, will return an empty string.
     * @return a text representation of the node.
     */
    @NotNull
    public static String nodeToString(AST node) {
        if (node == null) {
            return "";
        }
        if (node.getNumberOfChildren() > 0) {
            StringBuilder sb = new StringBuilder();

            AST child = node.getFirstChild();
            sb.append(nodeToString(child));
            sb.append(node.getText());

            do {
                child = child.getNextSibling();
                sb.append(nodeToString(child));
            } while (child != null);

//            sb.append("["+nodeToString(node.getNextSibling())+"]");
            return sb.toString();
        } else {
            return node.getText();
        }
    }

    /**
     * Returns the count of the nodes of type 'include' in the AST tree whose root is 'node'. Does not advance into
     * nodes of the type 'exclude'.
     *
     * @param node    the root node where to begin the search. If null, the method will return 0.
     * @param include the type of the nodes to count.
     * @param exclude the types of the nodes into which the algorithm should not dig into.
     * @return the total count of nodes of type 'include' under 'node'.
     */
    public static int getNodeCount(AST node, int include, int... exclude) {
        if (node == null) {
            return 0;
        }
        for (int ex : exclude) {
            if (node.getType() == ex) {
                return 0;
            }
        }
        int count = 0;
        AST child = node.getFirstChild();
        while (child != null) {
            count += getNodeCount(child, include, exclude);
            child = child.getNextSibling();
        }
        if (node.getType() == include) {
            count++;
        }
        return count;
    }
}
