// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.util;

import org.jetbrains.annotations.NotNull;

import java.io.*;
import java.util.List;

/**
 * Library class of useful file operations.
 *
 * @author Esko Luontola
 */
public class FileTools {

    /**
     * Reads the given file to a string.
     *
     * @param file the file to be read.
     * @return the contents of the file as a string.
     * @throws java.io.IOException  if the file can not be read.
     * @throws NullPointerException if file is null.
     */
    @NotNull
    public static String readFile(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException();
        }
        Reader in = new FileReader(file);
        StringBuilder sb = new StringBuilder((int) file.length());
        int i;
        while ((i = in.read()) != -1) {
            sb.append((char) i);
        }
        in.close();
        return sb.toString();
    }

    /**
     * Adds recursively all files in the specified directory to a list. Will not add duplicates to the result list.
     *
     * @param result where all files that were found will be added.
     * @param file   the directory which to read recursively. If it is a regular file, it will be added to the result
     *               list.
     * @throws NullPointerException if any parameter is null.
     */
    public static void recursiveAdd(List<File> result, File file) {
        if (file.isDirectory()) {
            File[] subdirFiles = file.listFiles();
            if (subdirFiles == null) {
                System.err.println("Unable to read directory: " + file);
                return;
            }
            for (File f : subdirFiles) {
                recursiveAdd(result, f);
            }
        } else if (!result.contains(file)) {
            result.add(file);
        }
    }

    public static void recursiveCopy(File from, File toDir, boolean overwrite) throws IOException {
        if (from.equals(toDir)) {
            throw new IOException("\"" + from + "\" and \"" + toDir + "\" are the same file");
        } else if (!from.exists()) {
            throw new IOException("No such file: " + from);
        }
        if (from.isDirectory()) {
            File[] files = from.listFiles();
            toDir.mkdirs();
            for (File file : files) {
                recursiveCopy(file, new File(toDir, from.getName()), overwrite);
            }
        } else {
            File toFile = new File(toDir, from.getName());
            if (!overwrite && toFile.exists()) {
                throw new IOException("Output file exists: " + toFile);
            }
            toDir.mkdirs();
            copy(from, toFile);
        }
    }

    private static void copy(File src, File dst) throws IOException {
        InputStream in = new FileInputStream(src);
        OutputStream out = new FileOutputStream(dst);

        // Transfer bytes from in to out
        byte[] buf = new byte[1024];
        int len;
        while ((len = in.read(buf)) > 0) {
            out.write(buf, 0, len);
        }
        in.close();
        out.close();
    }

    public static void recursiveDelete(File file) {
        if (file.isDirectory()) {
            File[] files = file.listFiles();
            for (File f : files) {
                recursiveDelete(f);
            }
            file.delete();
        } else if (file.isFile()) {
            file.delete();
        }
    }
}
