// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.util.commandline;

import org.jetbrains.annotations.*;

import java.text.ParseException;
import java.util.*;

/**
 * Parser for handling command line parameters. Keeps list of the available options and arguments that can be passed on
 * to a command line program. Parses the command line parameters and executes the actions that have been specified for
 * the supplied parameters.
 *
 * @author Esko Luontola
 */
public class CommandLineParser {

    /**
     * Argument name reserved for the command line options.
     */
    public static final String OPTION_ARGUMENT_NAME = "OPTION";

    public static final int DOCUMENTATION_COL1 = 2;
    public static final int DOCUMENTATION_COL2 = 28;
    public static final int DOCUMENTATION_WIDTH = 80;

    @NotNull
    private String command;

    @NotNull
    private String description;

    @NotNull
    private List<Option> options;

    @NotNull
    private List<Argument> arguments;

    @NotNull
    private Option helpOption;

    /**
     * Creates a new parser for the command line parameters.
     *
     * @param command     the name of the command that will start the program. Any whitespace will be trimmed from the
     *                    beginning and end of the string.
     * @param description a short description of what the program does.
     * @throws NullPointerException     if command is null.
     * @throws IllegalArgumentException if command contains only whitespace characters.
     */
    public CommandLineParser(String command, String description) {
        if (command == null) {
            throw new NullPointerException("command is null");
        } else if (description == null) {
            throw new NullPointerException("description is null");
        }
        command = command.trim();
        if (command.length() == 0) {
            throw new IllegalArgumentException("The command must contain at least 1 non-whitespace character.");
        }
        description = description.trim();
        this.command = command;
        this.description = description;
        this.options = new ArrayList<Option>();
        this.arguments = new ArrayList<Argument>();
        this.helpOption = new HelpOption(this);
    }

    /**
     * Adds a new option to the available command line options. The options will be listed by the --help command in the
     * order that they were added. Two options may not have such names, that they could be mistaken for each other.
     *
     * @param option the option to be added to available command line options.
     * @throws NullPointerException     if option is null.
     * @throws IllegalArgumentException if the new option could be mistaken for an existing option, because their
     *                                  command names are similar.
     */
    public void addOption(Option option) {
        if (option == null) {
            throw new NullPointerException();
        }

        // make sure that this new option does not overlap with an existing option
        String longName = option.getLongName();
        String shortName = option.getShortName();
        for (Option o : options) {

            // check overlapping long names
            if (longName != null && longName.equals(o.getLongName())) {
                throw new IllegalArgumentException(
                        "The option --" + longName + " overlaps with an existing option: --" + o.getLongName());
            }

            // check overlapping short names
            String s = o.getShortName();
            if (shortName == null || s == null) {
                continue;
            }
            for (int i = 1; i <= shortName.length(); i++) {
                if (s.startsWith(shortName.substring(0, i))) {
                    throw new IllegalArgumentException(
                            "The option -" + shortName + " overlaps with an existing option: -" + s);
                }
            }
        }
        this.options.add(option);
    }

    /**
     * Adds the default help option (--help) to the option list.
     *
     * @throws IllegalArgumentException if an option already uses the "--help" switch.
     */
    public void addDefaultHelpOption() {
        addOption(helpOption);
    }

    /**
     * Adds a new argument to the availabel command line parameters. The arguments will need to be entered on command
     * line in the order that they were added. Two arguments may not have the same names.
     *
     * @param argument the argument to be added to command line parameters.
     * @throws NullPointerException     if argument is null.
     * @throws IllegalArgumentException if the new argument has the same name as an existing argument.
     */
    public void addArgument(Argument argument) {
        if (argument == null) {
            throw new NullPointerException();
        }
        String name = argument.getName();
        if (name.equals(OPTION_ARGUMENT_NAME)) {
            throw new IllegalArgumentException(
                    "There is already an argument with the same name: " + name);
        }
        for (Argument a : arguments) {
            if (name.equals(a.getName())) {
                throw new IllegalArgumentException(
                        "There is already an argument with the same name: " + name);
            }
        }
        this.arguments.add(argument);
    }

    /**
     * Parses the command line parameters and executes the code for each option and argument that has been specified. If
     * the command line parameters were correct, the method will close normally. If there were errors in the parameters,
     * an exception with the explanation will be thrown.
     *
     * @param params the command line parameters as they would be passed on to a <code>main(String[] args)</code>
     *               method.
     * @throws NullPointerException if args is null.
     * @throws ParseException       if there is an unknown parameter. The message of the exception will contain the
     *                              reason.
     */
    public void parse(String[] params) throws ParseException {
        if (params == null) {
            throw new NullPointerException();
        }
        int i = 0;
        if (options.size() > 0) {

            // parse options
            for (; i < params.length; i++) {
                String param = params[i];

                if (param.startsWith("--")) {
                    // long option name
                    param = param.substring(2);             // remove prefixed "--"

                    Option option = findLongOption(param);
                    if (option == null) {
                        throw new ParseException("unrecognized option `" + params[i] + "'", i);
                    }

                    // get the argument part and check the option's argument rules
                    String argument = null;
                    if (param.indexOf('=') >= 0) {
                        argument = param.substring(param.indexOf('=') + 1);
                    }
                    if (option.isArgumentRequired() && argument == null) {
                        throw new ParseException("missing required argument `" + params[i] + "'", i);
                    } else if (!option.isArgumentSet() && argument != null) {
                        throw new ParseException("should not use an argument `" + params[i] + "'", i);
                    }
                    option.onSpecified(argument);

                } else if (param.startsWith("-")) {
                    // short option name
                    param = param.substring(1);             // remove prefixed "-"

                    // many short options may be put together, so continue until all of them are processed
                    while (param.length() > 0) {
                        Option option = findShortOption(param);
                        if (option == null) {
                            throw new ParseException("unrecognized option `-" + param + "'", i);
                        }
                        String shortName = option.getShortName();
                        assert shortName != null;

                        // get the argument part and check the option's argument rules
                        String argument = null;
                        if (param.equals(shortName)
                                && (i + 1) < params.length
//                              && !params[i + 1].startsWith("-") // HACK: caused problems with this command: -P -c
                                ) {
                            // this is the last one of the grouped options and the next parameter is an argument
                            argument = params[i + 1];
                        }
                        if (option.isArgumentRequired() && argument == null) {
                            throw new ParseException("missing required argument `-" + shortName + "'", i);
                        }
                        option.onSpecified(argument);

                        // process the next part of param
                        param = param.substring(shortName.length());
                        if (option.isArgumentSet() && argument != null) {       // skip the argument (avoid mistaking it for an option)
                            i++;
                        }
                    }

                } else {
                    // not an option, start processing the arguments
                    break;
                }
            }
        }

        // parse arguments
        for (int j = 0; j < arguments.size(); j++) {
            Argument argument = arguments.get(j);

            // calculate how many arguments still will be required
            int requiredArgumentsLeft = 0;
            for (int k = j + 1; k < arguments.size(); k++) {
                if (arguments.get(k).isRequired()) {
                    requiredArgumentsLeft++;
                }
            }

            // use as many of the parameters for the current argument as possible
            try {
                List<String> argParams = new ArrayList<String>();
                if (argument.isRequired()) {        // must use at least one parameter
                    argParams.add(params[i++]);
                } else if (params.length - i > requiredArgumentsLeft) {     // use one parameter if there are enough parameters left
                    argParams.add(params[i++]);
                }
                while (argument.isMultiple() && params.length - i > requiredArgumentsLeft) {
                    // multiple parameters for this argument, use as many parameters as allowed
                    argParams.add(params[i++]);
                }
                argument.onSpecified(argParams.toArray(new String[argParams.size()]));

            } catch (ArrayIndexOutOfBoundsException e) {
                throw new ParseException("missing required argument `" + argument.getName() + "'", j);
            }
        }

        // check that there are no parameters left over
        if (i < params.length) {
            throw new ParseException("unrecognized argument `" + params[i] + "'", i);
        }
    }

    @Nullable
    private Option findLongOption(String param) {

        // separate the option name part
        if (param.indexOf('=') >= 0) {
            param = param.substring(0, param.indexOf('='));         // remove trailing "=ARGUMENT"
        }

        // find the matching option
        for (Option option : options) {
            if (param.equals(option.getLongName())) {
                return option;
            }
        }

        // as a last resort try if the default help option matches
        if (param.equals(helpOption.getLongName())) {
            return helpOption;
        }
        return null;
    }

    @Nullable
    private Option findShortOption(String param) {

        // find the matching option
        for (Option option : options) {
            if (option.getShortName() == null) {
                continue;
            }

            // many short options may be put together, so try to find a substring match
            for (int i = 1; i <= param.length(); i++) {
                String s = param.substring(0, i);
                if (s.equals(option.getShortName())) {
                    return option;
                }
            }
        }
        return null;
    }

    /**
     * Returns the name of the command that will start the program.
     */
    @NotNull
    public String getCommand() {
        return command;
    }

    /**
     * Returns a short description of what the program does.
     */
    @NotNull
    public String getDescription() {
        return description;
    }

    /**
     * Returns the program documentation with minimal usage instructions.
     */
    @NotNull
    public String getQuickHelp() {
        StringBuilder sb = new StringBuilder();

        sb.append("Usage: ").append(getCommand());
        if (options.size() > 0) {
            sb.append(" [").append(OPTION_ARGUMENT_NAME).append("]");
        }
        if (options.size() > 1) {
            sb.append("...");
        }
        for (Argument argument : arguments) {
            sb.append(' ');
            if (argument.isRequired()) {
                sb.append(argument.getName());
            } else {
                sb.append("[").append(argument.getName()).append("]");
            }
            if (argument.isMultiple()) {
                sb.append("...");
            }
        }
        sb.append('\n');

        String[] words = getDescription().split("\\s");
        int lineLength = 0;
        for (String s : words) {
            if (lineLength + 1 + s.length() > DOCUMENTATION_WIDTH) {
                sb.append('\n');
                lineLength = 0;
            }
            if (lineLength > 0) {
                sb.append(' ');
            }
            sb.append(s);
            lineLength += 1 + s.length();
        }
        sb.append('\n');

        return sb.toString();
    }

    /**
     * Returns the program documentation with all the command line parameters.
     */
    @NotNull
    public String getHelp() {
        StringBuilder sb = new StringBuilder(getQuickHelp());
        sb.append('\n');
        for (Option option : options) {
            if (option.isArgumentRequired()) {
                sb.append("Mandatory arguments to long options are mandatory for short options too.\n");
                break;
            }
        }
        for (Option option : options) {
            sb.append(option.toString(DOCUMENTATION_COL1, DOCUMENTATION_COL2, DOCUMENTATION_WIDTH));
        }
        return sb.toString();
    }


    public static void main(String[] args) throws ParseException {
        CommandLineParser p = new CommandLineParser("ls",
                "List information about the FILEs (the current directory by default).\n" +
                        "Sort entries alphabetically if none of -cftuSUX nor --sort.");
        p.addOption(new Option("a", "all", "do not hide entries starting with .") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option("A", "almost-all", "do not list implied . and ..") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);

            }
        });
        p.addOption(new Option(null, "author", "print the author of each file") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option("b", "escape", "print octal escapes for nongraphic characters") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option(null, "block-size", "SIZE", true, "use SIZE-byte blocks") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option(null, "color", "WHEN", false,
                "control whether color is used to distinguish file types. WHEN may be `never', `always', or `auto'") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option("H", "dereference-command-line", "follow symbolic links listed on the command line") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option("I", "ignore", "PATTERN", true, "follow symbolic links listed on the command line") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option(null, "dereference-command-line-symlink-to-dir",
                "follow each command line symbolic link that points to a directory") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addOption(new Option("L", "dereference",
                "when showing file information for a symbolic link, show information for the file the link references rather than for the link itself") {
            public void onSpecified(String arg) {
                System.out.println(getShortName() + ":" + getLongName() + ":" + arg);
            }
        });
        p.addArgument(new Argument("TARGET", false, true) {
            public void onSpecified(String[] args) throws ParseException {
                System.out.println(getName() + ":" + args.length + ":" + Arrays.asList(args));
            }
        });
        p.addArgument(new Argument("DIRECTORY_PATH", true) {
            public void onSpecified(String[] args) throws ParseException {
                System.out.println(getName() + ":" + args.length + ":" + Arrays.asList(args));
            }
        });
        p.parse(new String[]{"-aAb",
                "--color",
                "--color=auto",
                "-I",
                "*+1",
                "--block-size=1234",
                "--block-size=",
                "-I",
                "*+1",
                "some1",
                "some2",
                "some3",
                "some4",
                "some5",
                "some6",
                "some7"
        });

        System.out.println();
        System.out.print(p.getHelp());
    }
}
