// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.util.commandline;

import org.jetbrains.annotations.*;

import java.text.ParseException;

/**
 * Represents one command line option that a command line application can take.
 *
 * @author Esko Luontola
 */
public abstract class Option {

    @Nullable
    private String shortName;

    @Nullable
    private String longName;

    @Nullable
    private String argumentName;

    private boolean argumentRequired;

    @NotNull
    private String description;


    /**
     * Creates a new command line option that takes no arguments.
     *
     * @param shortName   the short version of the option, without the prefixed "-". May contain upper and lowercase
     *                    alphabetical characters. To leave unspecified, leave it null, in which case the long version
     *                    must be specified.
     * @param longName    the long version of the option, without the prefixed "--". May contain lowercase alphabetical
     *                    characters and dash (-). To leave unspecified, leave it null, in which case the short version
     *                    must be specified.
     * @param description description of what the option does, for documentation purposes.
     * @throws IllegalArgumentException if the parameters do not follow the rules set for them.
     * @throws NullPointerException     if description is null.
     */
    public Option(String shortName, String longName, String description) {
        this(shortName, longName, null, false, description);
    }

    /**
     * Creates a new command line option.
     *
     * @param shortName        the short version of the option, without the prefixed "-". May contain upper and
     *                         lowercase alphabetical characters. To leave unspecified, leave it null, in which case the
     *                         long version must be specified.
     * @param longName         the long version of the option, without the prefixed "--". May contain lowercase
     *                         alphabetical characters and dash (-). To leave unspecified, leave it null, in which case
     *                         the short version must be specified.
     * @param argumentName     the name of the option's arguments, for documentation purposes. May contain uppercase
     *                         alphabetical characters. If an argument is specified, then the option must also have a
     *                         long name. To make the option take no arguments, leave it null.
     * @param argumentRequired true if the option will always take an argument, or false if there is no argument or it
     *                         is optional.
     * @param description      description of what the option does, for documentation purposes.
     * @throws IllegalArgumentException if the parameters do not follow the rules set for them.
     * @throws NullPointerException     if description is null.
     */
    public Option(String shortName, String longName, String argumentName, boolean argumentRequired,
                  String description) {
        if (shortName == null && longName == null) {
            throw new IllegalArgumentException(
                    "At least one of shortName and longName must be set.");
        } else if (shortName != null && !shortName.matches("[a-zA-Z]+")) {
            throw new IllegalArgumentException(
                    "shortName must contain one or more alphabetical characters (a-zA-Z) and nothing else. It was: " + shortName);
        } else if (longName != null && !longName.matches("[a-z][a-z-]*")) {
            throw new IllegalArgumentException(
                    "longName must contain only lowercase alphabetical characters (a-z) or dash (-) and must begin with an alphabet. It was: " + longName);
        } else if (argumentName == null && argumentRequired) {
            throw new IllegalArgumentException(
                    "When an argument is set to be required, you must also specify the argument's name.");
        } else if (argumentName != null && !argumentName.matches("[A-Z]+")) {
            throw new IllegalArgumentException(
                    "argumentName must contain one or more uppercase alphabetical characters (A-Z) and nothing else. It was: " + argumentName);
        } else if (argumentName != null && longName == null) {
            throw new IllegalArgumentException(
                    "When an argument is specified, then also longName must be specified.");
        } else if (description == null) {
            throw new NullPointerException("description is null");
        }
        this.shortName = shortName;
        this.longName = longName;
        this.argumentName = argumentName;
        this.argumentRequired = argumentRequired;
        this.description = description;
    }

    @Nullable
    public String getShortName() {
        return shortName;
    }

    @Nullable
    public String getLongName() {
        return longName;
    }

    @Nullable
    public String getArgumentName() {
        return argumentName;
    }

    /**
     * Returns true if this option can take an argument, otherwise false.
     */
    public boolean isArgumentSet() {
        return argumentName != null;
    }

    /**
     * Returns true if this option can take an argument and it is required, otherwise false.
     */
    public boolean isArgumentRequired() {
        return isArgumentSet() && argumentRequired;
    }

    @NotNull
    public String getDescription() {
        return description;
    }

    /**
     * Will return this option formatted as it would be by the --help command of a command line application. The
     * indentation of the first column is <code>CommandLineParser.DOCUMENTATION_COL1</code>, the second column is
     * <code>CommandLineParser.DOCUMENTATION_COL2</code> and line width is <code>CommandLineParser.DOCUMENTATION_WIDTH</code>.
     */
    @Override
    public String toString() {
        return this.toString(CommandLineParser.DOCUMENTATION_COL1,
                CommandLineParser.DOCUMENTATION_COL2,
                CommandLineParser.DOCUMENTATION_WIDTH);
    }

    /**
     * Will return this option formatted as it would be by the --help command of a command line application.
     *
     * @param col1  the beginning of the first column of text: the option and its arguments
     * @param col2  the beginning of the second column of text: the description
     * @param width the maximum width of a line, after which the text will be wrapped.
     * @return this option in a nicely formatted string.
     */
    public String toString(int col1, int col2, int width) {
        StringBuilder sb = new StringBuilder(width);

        // first column
        while (sb.length() < col1) {
            sb.append(' ');
        }

        // short name
        if (getShortName() != null) {
            sb.append("-").append(getShortName());
        } else {
            sb.append("  ");
        }
        if (getShortName() != null && getLongName() != null) {
            sb.append(", ");
        } else {
            sb.append("  ");
        }

        // long name
        if (getLongName() != null) {
            sb.append("--").append(getLongName());
        }

        // argument name (long name must be set when there are arguments)
        if (isArgumentSet()) {
            if (isArgumentRequired()) {
                sb.append("=").append(getArgumentName());
            } else {
                sb.append("[=").append(getArgumentName()).append("]");
            }
        }

        // description
        String[] descriptionWords = getDescription().split("\\s");
        int lineLength = sb.length();
        if (lineLength >= col2) {                           // put description on next line if the option names are very long
            sb.append('\n');
            lineLength = 0;
        }
        boolean wrapped = false;
        for (String word : descriptionWords) {

            // wrap at the end of line
            if (lineLength + 1 + word.length() >= width) {
                sb.append('\n');
                lineLength = 0;
                if (!wrapped) {                             // extra indentation for the second, third, etc. lines
                    col2 += 2;
                    wrapped = true;
                }
            }

            // second column
            while (lineLength < col2) {
                sb.append(' ');
                lineLength++;
            }

            // print description one word at a time
            sb.append(' ').append(word);
            lineLength += 1 + word.length();
        }

        // end with a line feed
        if (sb.charAt(sb.length() - 1) != '\n') {
            sb.append('\n');
        }
        return sb.toString();
    }

    /**
     * This method will be called by the CommandLineParser for each time when this option has been specified on the
     * command line. The subclasses should implement this method to do the actions that will reflect the option.
     *
     * @param arg the argument given to the option, or null if there are no arguments.
     * @throws ParseException if there is an unknown parameter. The message of the exception will contain the reason.
     */
    public abstract void onSpecified(String arg) throws ParseException;

}
