package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import net.orfjackal.util.FileTools;
import net.orfjackal.util.Tools;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Creates reports of the provided source code files.
 *
 * @author Esko Luontola
 */
public class AnalyzerBatch {

    @NotNull
    private List<FileAnalyzer> files;

    @NotNull
    private Set<File> classpaths;

    @Nullable
    private List<FileAnalyzer> classpathFileCache;

    private boolean countCodeLines;

    private boolean countCommentLines;

    private boolean countEmptyLines;

    private boolean includeOOMetrics;

    public AnalyzerBatch() {
        files = new ArrayList<FileAnalyzer>();
        classpaths = new HashSet<File>();
        classpathFileCache = null;
        countCodeLines = true;
        countCommentLines = false;
        countEmptyLines = false;
        includeOOMetrics = false;
    }

    /**
     * Adds a file to the list files being analyzed. Will read contents of the file to memory.
     *
     * @param file a source code file.
     * @throws NullPointerException if any parameter is null.
     * @throws IOException          if the parameter does not exist, is not a file or can not be read. The message of
     *                              the exception will contain the explanation.
     */
    public void addFile(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException();
        } else if (!file.exists()) {
            throw new IOException("No such file: " + file);
        } else if (!file.isFile()) {
            throw new IOException("Not a file: " + file);
        } else if (!file.canRead()) {
            throw new IOException("File can not be read: " + file);
        }
        FileAnalyzer fa = new FileAnalyzer(this, file);
        files.add(fa);
        classpathFileCache = null;
    }

    /**
     * Updates the classpath list by extracting the root folder from the files being analyzed.
     */
    private void updateClasspaths() {

        // extract the classpaths from the files
        for (FileAnalyzer fa : files) {
            String[] dirs = fa.getPackage().split("\\.");
            try {
                File f = fa.getFile().getCanonicalFile();
                for (int i = 0; i < dirs.length + 1 && f != null; i++) {
                    f = f.getParentFile();
                }
                if (f != null) {
                    classpaths.add(f);      // the Set will remove all duplicates
                }
            } catch (IOException e) {
                if (Tools.isDebug()) e.printStackTrace();
            }
        }
    }

    /**
     * Returns an array of the directories in the classpath. Updates the classpath list from the files that are being
     * analyzed.
     */
    public File[] getClasspaths() {
        updateClasspaths();
        return classpaths.toArray(new File[classpaths.size()]);
    }

    public boolean isCountCodeLines() {
        return countCodeLines;
    }

    public void setCountCodeLines(boolean countCodeLines) {
        this.countCodeLines = countCodeLines;
    }

    public boolean isCountCommentLines() {
        return countCommentLines;
    }

    public void setCountCommentLines(boolean countCommentLines) {
        this.countCommentLines = countCommentLines;
    }

    public boolean isCountEmptyLines() {
        return countEmptyLines;
    }

    public void setCountEmptyLines(boolean countEmptyLines) {
        this.countEmptyLines = countEmptyLines;
    }

    public boolean isIncludeOOMetrics() {
        return includeOOMetrics;
    }

    public void setIncludeOOMetrics(boolean includeOOMetrics) {
        this.includeOOMetrics = includeOOMetrics;
    }

    /**
     * Returns a report which contains the line counts of each of the input files. For each file is reported how many
     * lines of code it includes (according to the set LOC definition) and the name of the file. The total sum of the
     * line counts is told on the last line.
     * <p/>
     * The report will end with the line freed character '\n'.
     */
    @NotNull
    public String getLineCountReport() {
        StringBuilder sb = new StringBuilder();
        int totalLines = 0;
        for (FileAnalyzer file : files) {
            int lines = getLineCount(file);
            totalLines += lines;
            sb.append(lines);
            sb.append('\t');
            sb.append(file.getFile().getPath());
            sb.append('\n');
        }
        sb.append(totalLines);
        sb.append("\ttotal\n");
        return sb.toString();
    }

    private int getLineCount(FileAnalyzer file) {
        int count = 0;
        if (isCountCodeLines()) {
            count += file.getCodeLines();
        }
        if (isCountCommentLines()) {
            count += file.getCommentLines();
        }
        if (isCountEmptyLines()) {
            count += file.getEmptyLines();
        }
        return count;
    }

    /**
     * Returns a report which contains a structural analysis of each of the input files. The report will contain
     * <ul><li>the name of the source code file(s) processed</li> <li>for each source file:</li> <ul><li>the name of the
     * package the file belongs to,</li><li>the number and names of packages imported,</li><li>the number and names of
     * classes in the file,</li></ul> <li>for each class:</li> <ul><li>the number and names of the methods in the
     * class,</li></ul> <li>for each method:</li> <ul><li>the number of if and switch statements,</li><li>the number of
     * for, while and do-while statements.</li></ul></ul> The report will end with the line freed character '\n'.
     *
     * @throws ANTLRException if there was a parse exception in processing one of the files.
     */
    @NotNull
    public String getStructureReport() throws ANTLRException {
        StringBuilder sb = new StringBuilder();
        for (FileAnalyzer file : files) {
            sb.append(file.getFile().getPath());
            sb.append('\n');
            try {
                sb.append(file.getStructureReport());
                sb.append('\n');
            } catch (ANTLRException e) {
                if (Tools.isDebug()) e.printStackTrace();
                throw new ANTLRException("Parse exception in file '" + file.getFile() + "': " + e.getMessage());
            }
        }
        return sb.toString();
    }

    /**
     * Tries to find a the specified class from the classpath.
     *
     * @param classPackage the name of the package of the class.
     * @param className    the name of the class.
     * @return the associated ClassAnalyzer, or null if it was not found.
     * @throws NullPointerException if any argument is null.
     */
    @Nullable
    public ClassAnalyzer findClass(String classPackage, String className) {
//        System.out.println("findClass: " + classPackage + "." + className); // DEBUG
        if (classPackage == null) {
            throw new NullPointerException();
        } else if (className == null) {
            throw new NullPointerException();
        }

        // look for the class in the files being analyzed
        for (FileAnalyzer fa : files) {
            ClassAnalyzer ca = fa.getClass(classPackage, className);
            if (ca != null) {
                return ca;
            }
        }

        // look for the class in other files in the classpath
        String fileSeparator = System.getProperty("file.separator");
        String classFile = classPackage.replace(".", fileSeparator) + fileSeparator + className + ".java";
        for (File classpath : getClasspaths()) {
            File f = new File(classpath, classFile);
            if (f.exists() && f.isFile() && f.canRead()) {
                try {
                    FileAnalyzer fa = new FileAnalyzer(this, f);
                    ClassAnalyzer ca = fa.getClass(classPackage, className);
                    if (ca != null) {
                        return ca;
                    }
                } catch (IOException e) {
                    if (Tools.isDebug()) e.printStackTrace();
                }
            }
        }
        return null;
    }

    @NotNull
    public List<FileAnalyzer> getClasspathFileCache() {
        if (classpathFileCache == null) {
            ArrayList<FileAnalyzer> result = new ArrayList<FileAnalyzer>();

            // go through all files in the classpaths
            List<File> files = new ArrayList<File>();
            for (File classpath : getClasspaths()) {
                FileTools.recursiceAdd(files, classpath);
            }
            for (File file : files) {

                // ignore non-Java files
                if (!file.getName().matches(".*?\\.java\\z")) {
                    continue;
                }

                // analyze all Java files and add them to the cache
                try {
                    result.add(new FileAnalyzer(this, file));
                } catch (IOException e) {
                    if (Tools.isDebug()) e.printStackTrace();
                }
            }
            classpathFileCache = result;
        }
        return classpathFileCache;
    }

    /**
     * Tries to find from the classpath all classes that extend or implement the specified class. Only classes that
     * inherit it directly are returned.
     *
     * @param parentClass the parent class, whose direct children we are looking for.
     * @return an array of the classes that were found. If none were found, the size of the array is zero.
     * @throws NullPointerException if any argument is null.
     */
    @NotNull
    public ClassAnalyzer[] findChildClasses(ClassAnalyzer parentClass) {
        if (parentClass == null) {
            throw new NullPointerException();
        }
        List<ClassAnalyzer> result = new ArrayList<ClassAnalyzer>();
        List<FileAnalyzer> files = getClasspathFileCache();

        // find the classes that extend or implement parentClass
        for (FileAnalyzer fa : files) {
            for (ClassAnalyzer ca : fa.getClasses()) {
                ClassAnalyzer caExtends = ca.getExtends();
                ClassAnalyzer[] caImplements = ca.getImplements();
                boolean isChild = false;
                if (caExtends != null && caExtends.equals(parentClass)) {
                    isChild = true;
                }
                for (ClassAnalyzer impl : caImplements) {
                    if (impl.equals(parentClass)) {
                        isChild = true;
                    }
                }
                if (isChild) {
                    result.add(ca);         // is a child of parentClass, add to results
                }
            }
        }
        return result.toArray(new ClassAnalyzer[result.size()]);
    }
}
