package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import net.orfjackal.util.commandline.CommandLineParser;
import net.orfjackal.util.commandline.FileArgument;
import net.orfjackal.util.commandline.Option;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;

/**
 * Handles the command line parameters that this program takes and executes the program.
 *
 * @author Esko Luontola
 */
public class SorsastusLauncher extends CommandLineParser {

    private static final int LOC_OPTION_CODE_LINES = 1;
    private static final int LOC_OPTION_COMMENT_LINES = 2;
    private static final int LOC_OPTION_EMPTY_LINES = 4;
    private static final int LOC_OPTION_TOTAL_LINES = 7;
    private static final int STRUCTURE_OPTION_OOMETRICS = 1;

    private int locOptions = 0;
    private int structureOptions = 0;
    private boolean doStructure = false;

    private AnalyzerBatch analyzer;

    public SorsastusLauncher(AnalyzerBatch analyzer) {
        super(Sorsastus.APP_COMMAND, Sorsastus.APP_DESCRIPTION);
        if (analyzer == null) {
            throw new NullPointerException("analyzer is null");
        }
        this.analyzer = analyzer;

        addOption(new Option("c", "code-lines",
                "include effective lines of code to the LOC count (default, if none of `-mets' defined)") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_CODE_LINES;
            }
        });
        addOption(new Option("m", "comment-lines", "include comment lines to the LOC count") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_COMMENT_LINES;
            }
        });
        addOption(new Option("e", "empty-lines", "include empty lines to the LOC count") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_EMPTY_LINES;
            }
        });
        addOption(new Option("t", "total-lines", "include all lines to the LOC count (the same as `-cme')") {
            public void onSpecified(String arg) {
                locOptions = locOptions | LOC_OPTION_TOTAL_LINES;
            }
        });
        addOption(new Option("s", "structure", "output a detailed report of the structure of the source code files") {
            public void onSpecified(String arg) {
                doStructure = true;
            }
        });
        addOption(new Option("o", "oo-metrics", "include object-oriented metrics to the structure report (slow)") {
            public void onSpecified(String arg) {
                structureOptions = structureOptions | STRUCTURE_OPTION_OOMETRICS;
            }
        });
        addDefaultHelpOption();
        addOption(new Option(null, "version", "output version information and exit") {
            public void onSpecified(String arg) {
                System.out.println(Sorsastus.APP_NAME + " " + Sorsastus.APP_VERSION);
                System.out.println(Sorsastus.COPYRIGHT);
                System.exit(0);
            }
        });

        addArgument(new FileArgument("FILE_OR_DIR", true, true) {
            public void onSpecified(File[] files) throws ParseException {
                for (File file : files) {
                    try {
                        SorsastusLauncher.this.analyzer.addFile(file);
                    } catch (IOException e) {
                        throw new ParseException(e.getMessage(), 0);
                    }
                }
            }
        });
    }

    @Override
    public void parse(String[] params) {
        try {
            super.parse(params);
        } catch (ParseException e) {
            System.out.println(Sorsastus.APP_COMMAND + ": " + e.getMessage());
            System.out.println("Try `" + Sorsastus.APP_COMMAND + " --help' for more information.");
            System.exit(1);
        }

        // set the options
        if (locOptions == 0) {
            locOptions = locOptions | LOC_OPTION_CODE_LINES;
        }
        analyzer.setCountCodeLines((locOptions & LOC_OPTION_CODE_LINES) != 0);
        analyzer.setCountCommentLines((locOptions & LOC_OPTION_COMMENT_LINES) != 0);
        analyzer.setCountEmptyLines((locOptions & LOC_OPTION_EMPTY_LINES) != 0);
        analyzer.setIncludeOOMetrics((structureOptions & STRUCTURE_OPTION_OOMETRICS) != 0);

        // print the requested report
        if (doStructure) {
            try {
                System.out.print(analyzer.getStructureReport());
            } catch (ANTLRException e) {
                System.out.println(Sorsastus.APP_COMMAND + ": " + e.getMessage());
                System.exit(1);
            }
        } else {
            System.out.print(analyzer.getLineCountReport());
        }
    }
}
