// Tietorakenteiden harjoitusty (syksy 2004), Esko Luontola
// Tietoliikenneverkon reititysalgoritmien vertailu, "Sim Router"

/**
 * Reitittimiss kytettv puskuroiva lhetysportti toiseen reitittimeen
 * lhteville paketeille. SendPort-luokka on simulaation aktiivinen osapuoli, 
 * jota ajastin kutsuu aina, kun tmn portin kautta on aika saada yksi paketti 
 * lhetetty. Tm puolestaan kutsuu lhettv reititint (tick(int) metodi)
 * ja vastaanottoporttia (push() metodi). Vastaanottoportti kutsuu vastaanottavaa 
 * reititint (tick() metodi).
 */
public class SendPort {
	
	/**
	 * Reititin joka antaa paketteja tlle portille.
	 */
	private final Router from;
	
	/**
	 * Vastaanottoportti jonne tm ohjaa pakettinsa.
	 */
	private final RecievePort to;
	
	/**
	 * Puskuri siirtoa odottaville paketeille.
	 */
	private final PacketBuffer buffer;
	
	/**
	 * Kuinka monta aikayksikk paketilta kest kulkea tmn portin lpi.
	 */
	private final int transferDelay;
	
	/**
	 * Siirrossa oleva paketti, joka ei ole tll hetkell lhetys- eik 
	 * vastaanottopuskurissa.
	 */
	private Packet currentTransfer = null;
	
	/**
	 * Mill ajanhetkell (aikayksikk simulaation alusta) siirrossa oleva
	 * paketti lhetetn eteenpin vastaanottajalle.
	 */
	private int currentTransferTimeout = 0;
	
	/**
	 * Luo lhetysportin reitittimen from ja vastaanottoportin to vlille, 
	 * puskurikoolla bufferSize. Paketin kulkeminen tmn portin kautta kest 
	 * transferDelay aikayksikk (0 tai suurempi). Vastaanottoporttia ei saa 
	 * olla liitetty toiseen lhetysporttiin, koska se ei tarkista mist saa 
	 * pakettinsa, ja voisi siten hyvksy mist tahansa tulevia paketteja.
	 */
	public SendPort(Router from, RecievePort to, int bufferSize, int transferDelay) {
		if (from == null || to == null) {
			throw new NullPointerException();
		}
		if (transferDelay < 0) {
			transferDelay = 0;
		}
		this.from = from;
		this.buffer = new PacketBuffer(bufferSize);
		this.to = to;
		this.transferDelay = transferDelay;
	}
	
	/**
	 * Tyhjent pakettipuskurin.
	 */
	public void flushPackets() {
		while (!buffer.isEmpty()) {
			buffer.pop();
		}
		currentTransfer = null;
		currentTransferTimeout = 0;
	}
	
	/**
	 * Kuinka kauan paketilta kest kulkea tmn portin lpi.
	 */
	public int getTransferDelay() {
		return transferDelay;
	}
	
	/**
	 * Kuinka kauan paketti joutuu odottamaan puskurissa, jos se laitetaan nyt
	 * jonoon.
	 */
	public int getQueueDelay() {
		if (currentTransfer == null) {
			return 0;
		} else {
			return (currentTransferTimeout - Ticker.getTime()) + (transferDelay * buffer.getCount());
		}
	}
	
	/**
	 * Asettaa ajanhetken jolloin meneilln oleva tiedonsiirto valmistuu.
	 */
	public void setTimeout(int tick) {
		currentTransferTimeout = tick;
	}
	
	/**
	 * Palauttaa ajanhetken jolloin meneilln oleva tiedonsiirto valmistuu.
	 */
	public int getTimeout() {
		return currentTransferTimeout;
	}
	
	/**
	 * Ajastin kutsuu tt metodia antaakseen suoritusvuoron tlle lhetysportille.
	 * Parametrina tmnhetkinen aikayksikk simulaation alusta. Tarkistaa onko
	 * meneilln oleva siirto valmis, jolloin tm lhett paketin eteenpin.
	 */
	public void tick(int tick) {
		if (currentTransfer != null && currentTransferTimeout <= tick) {
			completeTransfer();
		} else {
			System.out.println("SendPort.tick(tick="+tick+") : "
				+"Unnecessary method call (currentTransferExists=="+(currentTransfer != null)
				+", currentTransferTimeout=="+currentTransferTimeout+")");
		}
	}
	
	/**
	 * Siirt paketin puskurista lhetysvuoroon ja asettaa ajastimen.
	 * Jos paketti on pikalhetyksess, ajastinta ei aseteta, vaan kutsutaan
	 * vlittmsti completeTransfer() metodia.
	 */
	private void beginTransfer() {
		if (currentTransfer == null && !buffer.isEmpty()) {
			currentTransfer = buffer.pop();
			if (getTransferDelay() == 0) {
				completeTransfer();
			} else {
				Ticker.setTimeout(this, getTransferDelay());
			}
		}
	}
	
	/**
	 * Siirt paketin lhetysvuorosta vastaanottoporttiin ja kutsuu lopuksi 
	 * beginTransfer() metodia seuraavan paketin lhettmiseksi.
	 */
	private void completeTransfer() {
		if (currentTransfer != null) {
			currentTransfer.addTransferTime(getTransferDelay());
			to.push(currentTransfer);
			currentTransfer = null;
		}
		beginTransfer();
	}
	
	/**
	 * Kutsuu puskurin push() metodia. Jos puskuri on tysi, niin annettu paketti 
	 * hvitetn. 
	 *
	 * @see PacketBuffer#push(Packet)
	 */
	public void push(Packet packet) {
		if (buffer.isEmpty()) {
			if (!buffer.push(packet)) {
				packet.markAsDropped();
			}
			beginTransfer();
		} else {
			if (!buffer.push(packet)) {
				packet.markAsDropped();
			}
		}
	}
	
	/**
	 * Kutsuu puskurin pop() metodia. Ainoastaan tm olio saa kytt tt
	 * metodia lhettkseen paketin edelleen vastaanottopuskuriin!
	 *
	 * @see PacketBuffer#pop()
	 */
	private Packet pop() {
		return buffer.pop();
	}
	
	/**
	 * Kutsuu puskurin getCount() metodia.
	 *
	 * @see PacketBuffer#getCount()
	 */
	public int getCount() {
		return buffer.getCount();
	}
	
	/**
	 * Kutsuu puskurin isFull() metodia.
	 *
	 * @see PacketBuffer#isFull()
	 */
	public boolean isFull() {
		return buffer.isFull();
	}
	
	/**
	 * Kutsuu puskurin isEmpty() metodia.
	 *
	 * @see PacketBuffer#isEmpty()
	 */
	public boolean isEmpty() {
		return buffer.isEmpty();
	}
	
	/**
	 * Palauttaa vyln tilan esityksen muodossa:
	 * fromID (transferDelay) sendBuffer currentTransfer recieveBuffer toID
	 */
	public String toString() {
		StringBuffer s = new StringBuffer(50);
		s.append(from.getID());
		s.append(" ("+transferDelay+") ");
		s.append(buffer.toString());
		s.append(currentTransfer == null ? " [-] " : " [+] ");
		s.append(to.toString());
		return s.toString();
	}
}
