// Tietorakenteiden harjoitusty (syksy 2004), Esko Luontola
// Tietoliikenneverkon reititysalgoritmien vertailu, "Sim Router"

/**
 * Ker tilastoja verkkoon lhetetyist paketeista. Paketit rekisteridn 
 * registerPacket()-metodilla silloin, kun ne lhtetetn verkkoon, ja kun ne
 * poistuvat verkosta. Tilastoja kertn kerrallaan yhteen Statistics-olioon,
 * mit beginNew() ja endCurrent() -metodit hallinnoivat.
 */
public class Statistics {
	
	private static Statistics current = null;
	
	private int startTime;
	
	private int endTime;
	
	private int packetsSent = 0;
	
	private int packetsRecieved = 0;
	
	private int packetsDropped = 0;
	
	private long totalTransferTime = 0;
	
	private int minTransferTime = Integer.MAX_VALUE;
	
	private int maxTransferTime = 0;
	
	private long totalJumps = 0;
	
	private int minJumps = Integer.MAX_VALUE;
	
	private int maxJumps = 0;
	
	/**
	 * Luo uuden Statistics-olion, joka ker tilastoja tst hetkest alkaen.
	 */
	private Statistics() {
		startTime = Ticker.getTime();
		endTime = -1;
	}
	
	/**
	 * Aloittaa uusien tilastojen kermisen. Palauttaa luodun Statistics-olion.
	 */
	public static Statistics beginNew() {
		endCurrent();
		current = new Statistics();
		return current;
	}
	
	/**
	 * Lopettaa tilastojen kermisen. Palauttaa tmnhetkisen Statistics-olion
	 * tai null, jos tilastoja ei oltu kermss.
	 */
	public static Statistics endCurrent() {
		if (current != null) {
			Statistics s = current;
			current = null;
			s.endTime = Ticker.getTime();
			return s;
		} else {
			return null;
		}
	}
	
	/**
	 * Lis paketin tiedot tll hetkell kerttviin tilastoihin.
	 */
	public static void registerPacket(Packet packet) {
		if (current != null) {
			current.registerPacket2(packet);
		}
	}
	
	/**
	 * Lis paketin tiedot nihin tilastoihin.
	 */
	private void registerPacket2(Packet packet){
		switch (packet.getStatus()) {
		case Packet.TRANSFER:
			packetsSent++;
			break;
			
		case Packet.RECIEVED:
			packetsRecieved++;
			
			int transferTime = packet.getTransferTime();
			totalTransferTime += transferTime;
			if (transferTime > maxTransferTime) {
				maxTransferTime = transferTime;
			}
			if (transferTime < minTransferTime) {
				minTransferTime = transferTime;
			}
			
			int jumps = packet.getJumps();
			totalJumps += jumps;
			if (jumps > maxJumps) {
				maxJumps = jumps;
			}
			if (jumps < minJumps) {
				minJumps = jumps;
			}
			break;
			
		case Packet.DROPPED:
		case Packet.DROPPED_NO_ROUTE:
		case Packet.DROPPED_NO_TTL:
			packetsDropped++;
			break;
			
		default:
			break;
		}
	}
	
	/**
	 * Tilastojen ajanjakson pituus. Jos tilastoja kertn viel, palauttaa -1.
	 */
	public int getLength() {
		if (endTime < 0) {
			return -1;
		} else {
			return endTime - startTime;
		}
	}
	
	/**
	 * Lhetettyjen pakettien lukumr.
	 */
	public int getPacketsSent() {
		return packetsSent;
	}
	
	/**
	 * Kohteeseensa saapuneiden pakettien lukumr.
	 */
	public int getPacketsRecieved() {
		return packetsRecieved;
	}
	
	/**
	 * Tuhottujen pakettien lukumr.
	 */
	public int getPacketsDropped() {
		return packetsDropped;
	}
	
	/**
	 * Kadonneet paketit jaettuna kadonneiden ja saapuneiden summalla.
	 */
	public double getPacketLoss() {
		return packetsDropped / (double)(packetsRecieved + packetsDropped);
	}
	
	/**
	 * Saapuneiden pakettien keskimrinen siirtoaika.
	 */
	public double getAvgTransferTime() {
		return totalTransferTime / (double)(packetsRecieved);
	}
	
	/**
	 * Saapuneiden pakettien lyhyin siirtoaika.
	 */
	public int getMinTransferTime() {
		return minTransferTime;
	}
	
	/**
	 * Saapuneiden pakettien pisin siirtoaika.
	 */
	public int getMaxTransferTime() {
		return maxTransferTime;
	}
	
	/**
	 * Saapuneiden pakettien keskimrinen hyppyjen lukumr.
	 */
	public double getAvgJumps() {
		return totalJumps / (double)(packetsRecieved);
	}
	
	/**
	 * Saapuneiden pakettien pienin hyppyjen lukumr.
	 */
	public int getMinJumps() {
		return minJumps;
	}
	
	/**
	 * Saapuneiden pakettien suurin hyppyjen lukumr.
	 */
	public int getMaxJumps() {
		return maxJumps;
	}
	
	
	public String toString() {
		StringBuffer s = new StringBuffer();
		
		s.append("Length:\t"+getLength()+"\n");
		s.append("Packets Sent:\t"+getPacketsSent()+"\n");
		s.append("Packets Recieved:\t"+getPacketsRecieved()+"\n");
		s.append("Packets Dropped:\t"+getPacketsDropped()+"\n");
		s.append("Packet Loss:\t"+Math.round(getPacketLoss() * 100000)/1000.0+"%\n");
		
		s.append("Transfer Time min/avg/max:\t");
		s.append(getMinTransferTime()+"/"+Math.round(getAvgTransferTime() * 1000)/1000.0+"/"+getMaxTransferTime()+"\n");
		s.append("Jumps min/avg/max:\t");
		s.append(getMinJumps()+"/"+Math.round(getAvgJumps() * 1000)/1000.0+"/"+getMaxJumps()+"\n");
		
		return s.toString();
	}
}
